from functools import reduce
from operator import add

import click

from nbmanips import Notebook
from nbmanips.cli import export, get_selector

__all__ = [
    'cat',
]


@click.command(help='Concatenate Jupyter FILE(s) to standard output')
@click.argument('file', nargs=-1, required=True)
@click.option('--output', '-o', default=None)
@click.option(
    '--force',
    '-f',
    is_flag=True,
    default=False,
    help='Do not prompt for confirmation if file already exists',
)
@click.option(
    '--select',
    '-s',
    type=int,
    default=None,
    help='Notebook to apply selector on. if unused, selector will be applied to all notebooks',
)
def cat(file, select, output, force):
    nbs = [Notebook.read(notebook_path) for notebook_path in file]
    selector = get_selector()

    if select is None:
        nbs = [nb.select(selector) for nb in nbs]
    else:
        nbs[select] = nbs[select].select(selector)

    nb = reduce(add, nbs)
    if output:
        export(nb, ..., output, force=force)
    else:
        click.echo(nb.to_json())



import copy
import re
from functools import reduce

import click
import cloudpickle
from click import Group

from nbmanips.cli import get_selector
from nbmanips.selector import Selector
from nbmanips.selector.default_selector import DefaultSelector

__all__ = [
    'select',
]


def _is_digit(selector):
    match = re.fullmatch(r'(-?\d+)', selector)

    if match is None:
        match = re.fullmatch(r'\[(-?\d+)\]', selector)

    if match is None:
        return None
    return match.group(1)


def _is_slice(selector):
    match = re.fullmatch(r'((-?\d+))?(:(-?\d+)?)?(:(-?\d+)?)?', selector)

    if match is None:
        match = re.fullmatch(r'\[((-?\d+))?(:(-?\d+)?)?(:(-?\d+)?)?\]', selector)

    if match is None:
        return None

    ret = []
    for el in match.groups()[1::2]:
        if el is not None:
            el = int(el)
        ret.append(el)
    return ret


class SelectGroup(Group):
    @property
    def dynamic_commands(self):
        return DefaultSelector.default_selectors

    def resolve_command(self, ctx, args):
        cmd_name, cmd, new_args = super().resolve_command(ctx, args)
        if cmd is None:
            return cmd_name, cmd, new_args

        if cmd_name not in self.list_commands(ctx):
            cmd_name = 'INDEX/SLICE'
        elif cmd_name in self.commands:
            args = new_args

        return cmd_name, cmd, args

    def get_command(self, ctx, cmd_name):
        cmd = self.commands.get(cmd_name)

        if cmd is not None:
            return cmd

        if cmd_name in self.dynamic_commands:
            cmd = copy.deepcopy(select_unknown)

            select_func = self.dynamic_commands[cmd_name]
            short_description = select_func.__doc__.strip().split('\n')[0]
            cmd.help = short_description
        elif _is_slice(cmd_name):
            cmd = copy.deepcopy(select_unknown)
        elif cmd_name == '[INDEX/SLICE]':
            cmd = copy.deepcopy(select_unknown)
            cmd.help = 'Selects Cells based on their index'

        return cmd

    def list_commands(self, ctx):
        commands = set(self.commands)
        commands |= set(self.dynamic_commands)
        commands |= {'[INDEX/SLICE]'}
        return sorted(commands)


def get_params(ctx, **kwargs):
    return {
        'or_': ctx.parent.params['or_'] or ctx.params['or_'],
        'invert': ctx.parent.params['invert'] or ctx.params['invert'],
        'kwargs': kwargs,
    }


def select_params(func):
    decorators = [
        click.option('--or', '-o', 'or_', is_flag=True, default=False),
        click.option('--invert', '-i', is_flag=True, default=False),
        func,
    ]

    return reduce(lambda f, g: g(f), decorators[::-1])


@click.group(cls=SelectGroup)
@select_params
def select(**_):
    pass


@click.command()
@click.argument('selector', required=True)
@click.argument('arguments', nargs=-1, required=False)
@click.option('--kwarg', 'kwargs', multiple=True, type=(str, str))
@select_params
@click.pass_context
def select_unknown(ctx, selector, arguments, kwargs, **_):
    if selector == '[INDEX/SLICE]':
        raise ValueError('Provide an index. Example: nb select [-3:-1]')
    digit_match = _is_digit(selector)
    slice_match = _is_slice(selector)
    if digit_match:
        selector = int(digit_match)
    elif slice_match:
        selector = slice(*slice_match)

    params = get_params(ctx, **dict(kwargs))

    _select_unknown(selector, arguments, **params)


@select.command(
    name='has_output_type', help='Select cells that have a given output_type'
)
@click.argument('output_type', nargs=-1, required=True)
@select_params
@click.pass_context
def has_output_type(ctx, output_type, **_):
    arguments = [set(output_type)]

    params = get_params(ctx)
    _select_unknown('has_output_type', arguments, **params)


@select.command(name='has_slide_type', help='Select cells that have a given slide type')
@click.argument('slide_type', nargs=-1, required=True)
@select_params
@click.pass_context
def has_slide_type(ctx, slide_type, **_):
    arguments = [set(slide_type)]

    params = get_params(ctx)
    _select_unknown('has_slide_type', arguments, **params)


@select.command(name='contains', help='Selects Cells containing a certain text.')
@click.argument('text', type=str, required=True)
@click.option('--case', '-c', is_flag=True, default=False)
@click.option('--include-output', '-n', 'output', is_flag=True, default=False)
@click.option('--regex', '-r', is_flag=True, default=False)
@select_params
@click.pass_context
def contains(ctx, text, case, output, regex, **_):
    params = get_params(
        ctx,
        case=case,
        output=output,
        regex=regex,
    )

    _select_unknown('contains', [text], **params)


def _select_unknown(selector, arguments, kwargs, or_, invert):
    sel = Selector(selector, *arguments, **kwargs)
    if invert:
        sel = ~sel

    piped_selector = get_selector()
    if piped_selector is not None:
        sel = (piped_selector | sel) if or_ else (piped_selector & sel)

    click.echo(cloudpickle.dumps(sel), nl=False)



import json
import os
import zipfile

from nbmanips.notebook import Notebook


def _parent_directory(path: str):
    return os.path.abspath(os.path.join(path, os.pardir))


def _get_dirs(path, common_path):
    dirs = set()
    while path.startswith(common_path):
        rel_path = os.path.relpath(path, common_path)
        if rel_path in dirs:
            break
        dirs.add(rel_path)
        path = _parent_directory(path)
    return dirs


class DbcExporter:
    @staticmethod
    def _to_dbc_notebook(
        nb: Notebook, name=None, language=None, version='NotebookV1'
    ) -> dict:
        if version != 'NotebookV1':
            raise ValueError(f'Unsupported version: {version}')

        # json part
        notebook = {'version': version, 'commands': []}

        # name
        if name is not None:
            notebook['name'] = name
        elif nb.name is not None:
            notebook['name'] = nb.name

        # language
        language = language or nb.language
        if language is not None:
            notebook['language'] = language

        for cell in nb.iter_cells():
            source = cell.source
            if cell.type == 'markdown':
                source = '%md\n' + source

            jupyter = cell.metadata.get('jupyter', {})

            results = [
                output.to_html()
                for output in cell.outputs
                if output.output_type != 'error'
            ]
            errors = [
                output for output in cell.outputs if output.output_type == 'error'
            ]

            command = {
                'version': 'CommandV1',
                'commandTitle': cell.metadata.get('name', ''),
                'command': source,
                'results': None,
                'errorSummary': None,
                'error': None,
                'collapsed': cell.metadata.get('collapsed', False),
                'hideCommandCode': jupyter.get('source_hidden', False),
                'hideCommandResult': jupyter.get('outputs_hidden', False),
            }

            if results:
                content = '\n'.join(results)
                command['results'] = {
                    'type': 'html',
                    'data': f"<div class=\"ansiout\">{content}</div>",
                }

            if errors:
                error = errors[0]
                command[
                    'errorSummary'
                ] = f"<span class=\"ansired\">{error.ename}</span>: {error.evalue}"
                command[
                    'error'
                ] = f"<div class=\"ansiout\">{error.to_html()}\n{command['errorSummary']}</div>"

            notebook['commands'].append(command)

        return notebook

    def export(self, nb: Notebook, output_path: str, filename=None, **kwargs):
        dbc_nb = self._to_dbc_notebook(nb, **kwargs)
        filename = filename or f"{dbc_nb['name']}.{dbc_nb['language']}"
        with zipfile.ZipFile(output_path, mode='w') as zf:
            zf.writestr(filename, json.dumps(dbc_nb))

    @staticmethod
    def _check_common_path(file_list, common_path):
        if common_path is None:
            common_path = os.path.commonpath(
                [_parent_directory(path) for path in file_list]
            )
        else:
            common_path = os.path.abspath(common_path)
            if not os.path.isdir(common_path):
                raise ValueError(f'common_path ({common_path}) is not a directory')
            invalid = list(
                filter(lambda file: not file.startswith(common_path), file_list)
            )
            if invalid:
                raise ValueError(
                    f'files: {invalid} should start with common_path: {common_path}'
                )
        return common_path

    def write_dbc(self, file_list, output_path, common_path=None):
        # absolute paths
        file_list = [os.path.abspath(file) for file in file_list]
        common_path = self._check_common_path(file_list, common_path)

        dirs = set()
        with zipfile.ZipFile(output_path, mode='w') as zf:
            for file_path in file_list:
                dbc_nb = self._to_dbc_notebook(Notebook.read_ipynb(file_path))
                default_filename = (
                    f"{dbc_nb['name']}.{dbc_nb.get('language', 'python')}"
                )
                parent_path = _parent_directory(file_path)
                dirs |= _get_dirs(parent_path, common_path)
                zip_path = os.path.join(
                    os.path.relpath(parent_path, common_path), default_filename
                )
                zf.writestr(zip_path, json.dumps(dbc_nb))

            for directory in dirs:
                zip_info = zipfile.ZipInfo(directory + '/')
                content = {
                    'version': 'FolderV1',
                    'name': directory,
                    'guid': '',
                    'children': [],
                }
                zf.writestr(zip_info, json.dumps(content))



import click

import nbmanips.cli.collection as collection
import nbmanips.cli.explore as explore
import nbmanips.cli.transform as transform
from nbmanips import __version__
from nbmanips.cli.convert import convert
from nbmanips.cli.select import select

__all__ = ['nbmanips']


@click.group()
@click.version_option(__version__, prog_name='nbmanips')
def nbmanips():
    pass


def _add_commands(module):
    for command in module.__all__:
        command_name = command.strip('_').replace('_', '-')
        nbmanips.add_command(getattr(module, command), command_name)


nbmanips.add_command(convert)
nbmanips.add_command(select)
_add_commands(explore)
_add_commands(transform)
_add_commands(collection)


if __name__ == '__main__':
    nbmanips()



from typing import Callable

from .base_selectors import ListSelector, SelectorBase, TrueSelector


class Selector(SelectorBase):
    def __init__(self, *args, **kwargs):
        super().__init__()

    def __new__(cls, selector, *args, **kwargs):
        if callable(selector):
            from .callable_selector import CallableSelector

            return CallableSelector(selector, *args, **kwargs)
        elif isinstance(selector, int):
            from .index_selector import IndexSelector

            return IndexSelector(selector)
        elif isinstance(selector, str):
            from .default_selector import DefaultSelector

            return DefaultSelector(selector, *args, **kwargs)
        elif hasattr(selector, '__iter__'):
            return ListSelector(selector, *args, **kwargs)
        elif isinstance(selector, slice):
            assert not kwargs and not args

            from .slice_selector import SliceSelector

            return SliceSelector(selector)
        elif isinstance(selector, SelectorBase):
            return selector
        elif selector is None:
            return TrueSelector()
        raise ValueError(
            f'selector needs to be of type: (str, int, list, slice, NoneType): {type(selector)}'
        )

    def get_callable(self, nb: dict) -> Callable:
        raise NotImplementedError()


__all__ = ['SelectorBase', 'Selector']



from typing import Dict, Optional

from .cell_utils import _get_output_types, _to_html, total_size
from .output_parsers import ParserBase


class CellOutput:
    output_type = None
    _output_types: Dict[str, type] = {}
    _parsers: Dict[str, ParserBase] = {}

    def __init__(self, content):
        self.content = content

    @property
    def output_types(self) -> set:
        raise NotImplementedError()

    def to_str(self, parsers=None, parsers_config=None, excluded_data_types=None):
        return ''

    def to_html(self, excluded_data_types=None):
        return ''

    def erase_output(self, output_types: set):
        raise NotImplementedError()

    def has_output_type(self, output_types: set):
        """
        Select cells that have a given output_type

        :param output_types: Output Types(MIME types) to select: text/plain, image/png, ...
        :type output_types: set
        :return: a bool object (True if cell should be selected)
        """
        return output_types & self.output_types

    def byte_size(self, output_types: Optional[set]):
        if output_types is None or self.has_output_type(output_types):
            return total_size(self.content)
        return 0

    @classmethod
    def register_parser(cls, output_type, parser: ParserBase):
        cls._parsers[output_type] = parser

    @classmethod
    def get_parser(cls, parser_key: str, parser_config_dict: dict):
        if parser_key in cls._parsers:
            return cls._parsers[parser_key], parser_config_dict.get(parser_key, {})
        else:
            parser_key = parser_key.split('/')[0]
            return cls._parsers.get(parser_key, None), parser_config_dict.get(
                parser_key, {}
            )

    @property
    def default_parsers(self) -> set:
        return {key for key, value in self._parsers.items() if value.default_state}

    def __new__(cls, content, *args, **kwargs):
        output_type = content['output_type']
        output_class = cls._output_types[output_type]
        obj = super().__new__(output_class)
        output_class.__init__(obj, content, *args, **kwargs)
        return obj

    def __init_subclass__(cls, output_type=None, **kwargs):
        super().__init_subclass__(**kwargs)
        if output_type:
            cls._output_types[output_type] = cls
        cls.output_type = output_type


class StreamOutput(CellOutput, output_type='stream'):
    @property
    def text(self):
        return self.content['text']

    @property
    def output_types(self):
        return _get_output_types('text/plain')

    def to_str(self, parsers=None, parsers_config=None, excluded_data_types=None):
        parsers_config = parsers_config or {}
        excluded_data_types = (
            set() if excluded_data_types is None else set(excluded_data_types)
        )
        if self.output_types & excluded_data_types:
            return ''

        output_text = self.text
        if not isinstance(output_text, str):
            output_text = '\n'.join(output_text)

        parser, parser_config = self.get_parser('text/plain', parsers_config)
        if parser:
            return parser.parse(output_text, **parser_config)
        return output_text

    def to_html(self, excluded_data_types=None):
        excluded_data_types = (
            set() if excluded_data_types is None else set(excluded_data_types)
        )
        if self.output_types & excluded_data_types:
            return ''
        output_text = self.text
        if not isinstance(output_text, str):
            output_text = '\n'.join(output_text)
        return _to_html(output_text)

    def erase_output(self, output_types: set):
        return None if self.has_output_type(output_types) else self.content


class DataOutput(CellOutput):
    _default_data_types = ['text', 'image', 'text/html']

    @classmethod
    def _get_key(cls, dt, parsers):
        alt_dt = dt.split('/')[0]
        s1 = dt not in parsers and alt_dt not in parsers
        s2 = -(
            cls._default_data_types.index(dt)
            if dt in cls._default_data_types
            else (
                cls._default_data_types.index(alt_dt)
                if alt_dt in cls._default_data_types
                else -1
            )
        )
        return s1, s2

    def to_str(self, parsers=None, parsers_config=None, excluded_data_types=None):
        parsers = self.default_parsers if parsers is None else set(parsers)
        parsers_config = parsers_config or {}
        excluded_data_types = (
            set() if excluded_data_types is None else set(excluded_data_types)
        )

        data = self.content['data']
        for data_type in sorted(data, key=lambda x: self._get_key(x, parsers)):
            alt_data_types = _get_output_types(data_type)
            if alt_data_types & excluded_data_types:
                continue

            output_text = data[data_type]
            if not isinstance(output_text, str):
                output_text = '\n'.join(output_text)

            if alt_data_types & parsers:
                parser, parser_config = self.get_parser(data_type, parsers_config)
                if parser:
                    output_text = parser.parse(output_text, **parser_config)
            return output_text
        return ''

    def to_html(self, excluded_data_types=None):
        excluded_data_types = (
            set() if excluded_data_types is None else set(excluded_data_types)
        )

        data = self.content['data'].copy()
        for data_type in list(data.keys()):
            if _get_output_types(data_type) & excluded_data_types:
                data.pop(data_type, None)

        if 'text/html' in data:
            output_text = data['text/html']
            if not isinstance(output_text, str):
                output_text = '\n'.join(output_text)
            return output_text

        for data_type in data:
            if data_type.startswith('image'):
                content = data['image/png']
                return f'<img src="data:image/png;base64, {content}"/>'

        if 'text/plain' in data:
            output_text = data['text/plain']
            if not isinstance(output_text, str):
                output_text = '\n'.join(output_text)
            return _to_html(output_text)

        return ''

    def erase_output(self, output_types: set):
        for key in output_types:
            self.content['data'].pop(key, None)

        if not self.content['data']:
            return None

        return self.content

    @property
    def output_types(self):
        return _get_output_types(self.content['data'])

    def has_output_type(self, output_types: set):
        return output_types & self.output_types

    def byte_size(self, output_types: Optional[set]):
        if output_types is None:
            return total_size(self.content)
        elif not self.has_output_type(output_types):
            return 0
        else:
            return total_size(
                {
                    key: value
                    for key, value in self.content['data'].items()
                    if key in output_types
                }
            )


class ErrorOutput(CellOutput, output_type='error'):
    @property
    def ename(self):
        return self.content['ename']

    @property
    def evalue(self):
        return self.content['evalue']

    @property
    def traceback(self):
        return self.content['traceback']

    @property
    def output_types(self):
        return _get_output_types('text/error')

    def to_str(self, parsers=None, parsers_config=None, excluded_data_types=None):
        parsers_config = parsers_config or {}
        excluded_data_types = (
            set() if excluded_data_types is None else set(excluded_data_types)
        )
        if self.output_types & excluded_data_types:
            return ''

        output_text = '\n'.join(self.traceback + [f'{self.ename}: {self.evalue}'])
        parser, parser_config = self.get_parser('text/error', parsers_config)
        if parser:
            return parser.parse(output_text, **parser_config)
        return output_text

    def to_html(self, excluded_data_types=None):
        excluded_data_types = (
            set() if excluded_data_types is None else set(excluded_data_types)
        )
        if self.output_types & excluded_data_types:
            return ''

        return _to_html('\n'.join(self.traceback + [f'{self.ename}: {self.evalue}']))

    def erase_output(self, output_types: set):
        return None if self.has_output_type(output_types) else self.content


class DisplayData(DataOutput, output_type='display_data'):
    pass


class ExecuteResult(DataOutput, output_type='execute_result'):
    @property
    def execution_count(self):
        return self.content.get('execution_count', None)



from pathlib import Path

import nbmanips.exporters as _nb_exporters
from nbmanips.notebook import DBC, IPYNB, ZPLN, Notebook

__version__ = (Path(__file__).parent / 'VERSION').read_text(encoding='utf-8').strip()

__all__ = ['DBC', 'IPYNB', 'ZPLN', 'Notebook', '__version__']


# -- Register Exporters --
Notebook.register_exporter('dbc', _nb_exporters.DbcExporter, exporter_type='nbmanips')



from functools import partial
from typing import Callable

from ..cell import Cell
from . import SelectorBase


class SliceSelector(SelectorBase):
    def __init__(self, selector):
        self._slice = selector
        super().__init__()

    def get_callable(self, nb: dict) -> Callable[[Cell], bool]:
        new_slice = self.__adapt_slice(self._slice, len(nb.get('cells', [])))
        return self.__get_slice_selector(new_slice)

    @staticmethod
    def __adapt_slice(old_slice, n_cells):
        start, stop, step = old_slice.start, old_slice.stop, old_slice.step
        if stop is not None and stop < 0:
            stop = stop + n_cells
        if start is not None and start < 0:
            start = start + n_cells

        new_slice = slice(start, stop, step)
        return new_slice

    @classmethod
    def __get_slice_selector(cls, selector: slice) -> Callable[[Cell], bool]:
        start, stop, step = selector.start, selector.stop, selector.step
        if step and step < 0:
            start, stop = stop + 1, start + 1

        selector_list = []
        if start is not None:
            selector_list.append(lambda cell: cell.num >= start)
        if stop is not None:
            selector_list.append(lambda cell: cell.num < stop)
        if step is not None:
            selector_list.append(lambda cell: (cell.num - start) % abs(step) == 0)
        return partial(cls.__get_multiple_selector, selector_list=selector_list)

    @staticmethod
    def __get_multiple_selector(cell, selector_list: list):
        return all(sel(cell) for sel in selector_list)



from .cell_output import CellOutput
from .cells import Cell, CodeCell, MarkdownCell, RawCell
from .output_parsers import HtmlParser, ImageParser, TextParser

CellOutput.register_parser('text', TextParser())
CellOutput.register_parser('text/html', HtmlParser())
CellOutput.register_parser('image', ImageParser())


__all__ = ['Cell', 'CellOutput', 'MarkdownCell', 'CodeCell', 'RawCell']



from typing import Callable

from nbmanips._utils import partial
from nbmanips.selector import SelectorBase


class CallableSelector(SelectorBase):
    def __init__(self, selector, *args, **kwargs):
        if args or kwargs:
            self._selector = partial(selector, *args, **kwargs)
        else:
            self._selector = selector
        super().__init__()

    def get_callable(self, nb: dict) -> Callable[..., bool]:
        return self._selector



import json
import os
import shutil
import textwrap
from copy import deepcopy
from pathlib import Path
from typing import Any, Union

try:
    import nbconvert
except ImportError:
    nbconvert = None

from nbmanips.cell.cell_utils import PYGMENTS_SUPPORTED
from nbmanips.notebook.utils import (
    dict_to_ipynb,
    get_ipynb_name,
    read_dbc,
    read_ipynb,
    read_zpln,
    write_ipynb,
)

from .notebook_base import NotebookBase

try:
    import pygments.util
    from pygments.lexers import get_lexer_by_name
except ImportError:
    pygments = None
    get_lexer_by_name = None


class ClassicNotebook(NotebookBase):
    def update_cell_metadata(self, key: str, value: Any):
        """
        Add metadata to the selected cells
        :param key: metadata key
        :param value: metadata value
        """
        for cell in self.iter_cells():
            value = deepcopy(value)
            cell.update_metadata(key, value)

    def erase(self):
        """
        Erase the content of the selected cells
        """
        for cell in self.iter_cells():
            cell.set_source([])

    def erase_output(self, output_types=None):
        """
        Erase the output content of the selected cells
        """
        for cell in self.iter_cells():
            cell.erase_output(output_types)

    def delete(self):
        """
        Delete the selected cells
        """
        self.raw_nb['cells'] = [cell.cell for cell in self.iter_cells(neg=True)]

    def keep(self):
        """
        Delete all the non-selected cells
        """
        self.raw_nb['cells'] = [cell.cell for cell in self.iter_cells()]

    def first(self):
        """
        Return the number of the first selected cell
        :return:
        """
        for cell in self.iter_cells():
            return cell.num

    def last(self):
        """
        Return the number of the last selected cell
        :return:
        """
        for cell in reversed(list(self.iter_cells())):
            return cell.num

    def copy(self, selection=True, crop=True):
        """
        Copy the notebook instance

        :param selection: keep selection
        :param crop: crop on selection
        :return: a new copy of the notebook
        """
        cp = self.__class__(self.raw_nb, self.name, validate=False, copy=True)
        if selection:
            cp._selector = self._selector
            if crop:
                cp.keep()
                cp = cp.reset_selection()
        return cp

    def split(self, *args):
        """
        Split the notebook based passed selectors (typically cell indexes)

        :param args:
        :return:
        """
        return self.copy().select(args, type='or').split_on_selection()

    def split_on_selection(self):
        """
        Split the notebook based on the selected cells
        :return:
        """
        cp = self.reset_selection()
        notebooks = []
        prev = 0
        for cell in self.iter_cells():
            if cell.num == prev:
                continue

            notebooks.append(cp[prev : cell.num].copy())

            prev = cell.num
        notebooks.append(cp[prev:].copy())
        return notebooks

    def list(self):
        """
        Return the numbers of the selected cells
        :return:
        """
        return [cell.num for cell in self.list_cells()]

    def count(self):
        """
        Return the numbers of the selected cells
        :return:
        """
        return len(self)


class SlideShowMixin(ClassicNotebook):
    def mark_slideshow(self):
        self.raw_nb['metadata']['celltoolbar'] = 'Slideshow'

    def set_slide(self):
        self.tag_slide('slide')

    def set_skip(self):
        self.tag_slide('skip')

    def set_subslide(self):
        self.tag_slide('subslide')

    def set_fragment(self):
        self.tag_slide('fragment')

    def set_notes(self):
        self.tag_slide('notes')

    def tag_slide(self, tag):
        assert tag in {'-', 'skip', 'slide', 'subslide', 'fragment', 'notes'}
        self.update_cell_metadata('slideshow', {'slide_type': tag})

    def max_cells_per_slide(self, n_cells=3, n_images=1):
        from nbmanips.selector.default_selector import (
            has_output_type,
            has_slide_type,
            is_new_slide,
        )

        cells_count = 0
        img_count = 0
        for cell in self.iter_cells():
            is_image = has_output_type(cell, 'image/png')
            if is_new_slide(cell):
                cells_count = 1
                img_count = 0
            elif has_slide_type(cell, {'skip', 'fragment', 'notes'}):
                # Don't count
                pass
            else:
                cells_count += 1
            if is_image:
                img_count += 1

            if (n_cells is not None and cells_count > n_cells) or (
                n_images is not None and img_count > n_images
            ):
                cell.update_metadata('slideshow', {'slide_type': 'subslide'})
                cells_count = 1
                img_count = 1 if is_image else 0

    def auto_slide(
        self,
        max_cells_per_slide=3,
        max_images_per_slide=1,
        *_,
        delete_empty=True,
        title_tags='h1, h2',
    ):
        from nbmanips.selector.default_selector import is_new_slide

        # Delete Empty
        if delete_empty:
            self.select('is_empty').delete()

        # Each title represents
        self.select('with_css_selector', title_tags).set_slide()

        # Create a new slide only
        for cell in reversed(list(self.iter_cells())):
            if cell.num > 0 and is_new_slide(
                self[cell.num - 1].first_cell()
            ):  # previous cell is a new slide
                cell.update_metadata('slideshow', {'slide_type': '-'})

        # Set max cells per slide
        self.max_cells_per_slide(max_cells_per_slide, max_images_per_slide)


class ExportMixin(NotebookBase):
    __exporters = {
        'nbconvert': {
            'html': nbconvert.HTMLExporter,
            'slides': nbconvert.SlidesExporter,
            'python': nbconvert.PythonExporter,
            'markdown': nbconvert.MarkdownExporter,
            'script': nbconvert.ScriptExporter,
        }
    }

    @classmethod
    def register_exporter(cls, exporter_name: str, exporter, exporter_type='nbconvert'):
        exporters = cls.__exporters.setdefault(exporter_type, {})
        exporters[exporter_name] = exporter

    @classmethod
    def get_exporter(cls, exporter_name, *args, exporter_type='nbconvert', **kwargs):
        return cls.__exporters[exporter_type][exporter_name](*args, **kwargs)

    def to_json(self):
        """
        returns notebook as json string.
        """
        return json.dumps(self.raw_nb)

    def to_notebook_node(self):
        """
        returns notebook as an nbformat NotebookNode
        """
        return dict_to_ipynb(self.raw_nb)

    def convert(self, exporter_name, path, *args, exporter_type='nbmanips', **kwargs):
        assert exporter_type in {'nbmanips', 'nbconvert'}
        if exporter_type == 'nbconvert':
            return self.nbconvert(exporter_name, path, *args, **kwargs)

        exporter = self.get_exporter(exporter_name, exporter_type=exporter_type)

        return exporter.export(self, path, *args, **kwargs)

    def nbconvert(self, exporter_name, path, *args, template_name=None, **kwargs):
        notebook_node = self.to_notebook_node()

        if template_name is not None:
            kwargs['template_name'] = template_name

        exporter = self.get_exporter(
            exporter_name, *args, exporter_type='nbconvert', **kwargs
        )

        (body, resources) = exporter.from_notebook_node(notebook_node)

        # Exporting result
        build_directory, file_name = os.path.split(path)
        writer = nbconvert.writers.files.FilesWriter(build_directory=build_directory)

        _, ext = os.path.splitext(file_name)
        if ext:
            resources.pop('output_extension')

        writer.write(body, resources, file_name)

    def to_html(
        self,
        path,
        exclude_code_cell=False,
        exclude_markdown=False,
        exclude_raw=False,
        exclude_unknown=False,
        exclude_input=False,
        exclude_output=False,
        **kwargs,
    ):
        """
        Exports a basic HTML document.

        :param path: path to export to
        :param exclude_code_cell: exclude code cells from all templates if set to True.
        :param exclude_markdown: exclude markdown cells from all templates if set to True.
        :param exclude_raw: exclude raw cells from all templates if set to True.
        :param exclude_unknown: exclude unknown cells from all templates if set to True.
        :param exclude_input: exclude input prompts from all templates if set to True.
        :param exclude_output: exclude code cell outputs from all templates if set to True.
        :param kwargs: exclude_input_prompt, exclude_output_prompt, ...
        """
        return self.nbconvert(
            'html',
            path,
            exclude_code_cell=exclude_code_cell,
            exclude_markdown=exclude_markdown,
            exclude_raw=exclude_raw,
            exclude_unknown=exclude_unknown,
            exclude_input=exclude_input,
            exclude_output=exclude_output,
            **kwargs,
        )

    def to_py(self, path, **kwargs):
        """
        Exports a Python code file.
        Note that the file produced will have a shebang of '#!/usr/bin/env python'
        regardless of the actual python version used in the notebook.

        :param path: path to export to
        """
        return self.nbconvert('python', path, **kwargs)

    def to_md(
        self,
        path,
        exclude_code_cell=False,
        exclude_markdown=False,
        exclude_raw=False,
        exclude_unknown=False,
        exclude_input=False,
        exclude_output=False,
        **kwargs,
    ):
        """
        Exports to a markdown document (.md)

        :param path: path to export to
        :param exclude_code_cell: exclude code cells from all templates if set to True.
        :param exclude_markdown: exclude markdown cells from all templates if set to True.
        :param exclude_raw: exclude raw cells from all templates if set to True.
        :param exclude_unknown: exclude unknown cells from all templates if set to True.
        :param exclude_input: exclude input prompts from all templates if set to True.
        :param exclude_output: exclude code cell outputs from all templates if set to True.
        :param kwargs: exclude_input_prompt, exclude_output_prompt, ...
        """
        return self.nbconvert(
            'markdown',
            path,
            exclude_code_cell=exclude_code_cell,
            exclude_markdown=exclude_markdown,
            exclude_raw=exclude_raw,
            exclude_unknown=exclude_unknown,
            exclude_input=exclude_input,
            exclude_output=exclude_output,
            **kwargs,
        )

    def to_slides(
        self, path, scroll=True, transition='slide', theme='simple', **kwargs
    ):
        """
        Exports HTML slides with reveal.js

        :param path: path to export to
        :param scroll: If True, enable scrolling within each slide
        :type scroll: bool
        :param transition: Name of the reveal.js transition to use.
        :type transition: none, fade, slide, convex, concave and zoom.
        :param theme: Name of the reveal.js theme to use.
        See https://github.com/hakimel/reveal.js/tree/master/css/theme
        :type theme: beige, black, blood, league, moon, night, serif, simple, sky, solarized, white
        :param kwargs: any additional keyword arguments to nbconvert exporter
        :type kwargs: exclude_code_cell, exclude_markdown, exclude_input, exclude_output, ...
        """
        return self.nbconvert(
            'slides',
            path,
            reveal_scroll=scroll,
            reveal_transition=transition,
            reveal_theme=theme,
            **kwargs,
        )

    def to_dbc(
        self, path, filename=None, name=None, language=None, version='NotebookV1'
    ):
        """
        Exports Notebook to dbc archive file

        :param path: path to export to
        :param filename: filename of the notebook inside archive (e.i. notebook.python)
        :param name: name of the notebook
        :param language: language of the notebook
        :param version: version of dbc file (default is NotebookV1)
        :return:
        """
        self.convert(
            'dbc',
            path,
            exporter_type='nbmanips',
            filename=filename,
            name=name,
            language=language,
            version=version,
        )

    def _get_pygments_lexer(self, use_pygments):
        if use_pygments:
            pygments_lexer = self.metadata.get('language_info', {}).get(
                'pygments_lexer', None
            )
            pygments_lexer = pygments_lexer or self.metadata.get(
                'language_info', {}
            ).get('name', None)
            pygments_lexer = pygments_lexer or self.metadata.get('kernelspec', {}).get(
                'language', None
            )
        else:
            pygments_lexer = None

        if pygments_lexer is None:
            return pygments_lexer

        if get_lexer_by_name is None:
            raise ModuleNotFoundError(
                'You need to install pygments first.\n pip install pygments'
            )

        try:
            if pygments_lexer in {'ipython3', 'python', 'py', 'python3', 'py3'}:
                from pygments.lexers.python import PythonLexer

                return PythonLexer()

            return get_lexer_by_name(pygments_lexer)
        except pygments.util.ClassNotFound:
            return None

    def to_str(
        self,
        width=None,
        exclude_output=False,
        use_pygments=None,
        style='single',
        border_color=None,
        parsers=None,
        parsers_config=None,
        excluded_data_types=None,
        truncate=None,
    ):
        use_pygments = PYGMENTS_SUPPORTED if use_pygments is None else use_pygments
        pygments_lexer = self._get_pygments_lexer(use_pygments)

        return '\n'.join(
            cell.to_str(
                width=width,
                exclude_output=exclude_output,
                use_pygments=use_pygments,
                pygments_lexer=pygments_lexer,
                style=style,
                color=border_color,
                parsers=parsers,
                parsers_config=parsers_config,
                excluded_data_types=excluded_data_types,
                truncate=truncate,
            )
            for cell in self.iter_cells()
        )

    def to_text(self, path, *args, **kwargs):
        """
        Exports to visual text format
        :param path: path to export to
        :param args:
        :param kwargs:
        :return:
        """
        content = self.to_str(*args, use_pygments=False, border_color=False, **kwargs)
        with open(path, 'w', encoding='utf-8') as f:
            f.write(content)

    def to_ipynb(self, path):
        """
        Export to ipynb file
        :param path: target path
        """
        write_ipynb(self.raw_nb, path)

    def show(
        self,
        width=None,
        exclude_output=False,
        use_pygments=None,
        style='single',
        border_color=None,
        parsers=None,
        parsers_config=None,
        excluded_data_types=None,
        truncate=None,
    ):
        """
        Show the selected cells
        :param width:
        :param style:
        :param use_pygments:
        :param border_color:
        :param exclude_output:
        :param parsers:
        :param parsers_config:
        :param excluded_data_types:
        """
        str_repr = self.to_str(
            width=width,
            use_pygments=use_pygments,
            exclude_output=exclude_output,
            style=style,
            border_color=border_color,
            parsers=parsers,
            parsers_config=parsers_config,
            excluded_data_types=excluded_data_types,
            truncate=truncate,
        )
        if str_repr:
            print(str_repr)

    @classmethod
    def read_ipynb(cls, path, name=None, validate=False):
        """
        Read ipynb file
        :param path: path to the ipynb file
        :param name: name of the Notebook
        :return: Notebook object
        """
        nb = read_ipynb(path)
        nb = cls(nb, name or get_ipynb_name(path), validate=validate, copy=False)

        nb._original_path = path

        return nb

    @classmethod
    def read_dbc(cls, path, filename=None, encoding='utf-8', name=None, validate=False):
        dbc_name, nb = read_dbc(path, filename=filename, encoding=encoding)
        nb = cls(nb, name or dbc_name, validate=validate, copy=False)

        nb._original_path = path

        return nb

    @classmethod
    def read_zpln(cls, path, encoding='utf-8', name=None, validate=False):
        zpln_name, nb = read_zpln(path, encoding=encoding)
        nb = cls(nb, name or zpln_name, validate=validate, copy=False)

        nb._original_path = path

        return nb

    @classmethod
    def read(cls, path, name=None, validate=False, **kwargs):
        readers = {
            '.ipynb': cls.read_ipynb,
            '.dbc': cls.read_dbc,
            '.zpln': cls.read_zpln,
        }

        if not Path(path).exists():
            raise FileNotFoundError(f'Could not find: {path}')

        ext = Path(path).suffix.lower()
        reader = readers.get(ext, None)
        if reader:
            return reader(path, name=name, validate=validate, **kwargs)

        for reader in readers.values():
            try:
                return reader(path, name=name, validate=validate, **kwargs)
            except Exception:
                continue

        raise ValueError('Could not determine the notebook type')


class NotebookMetadata(NotebookBase):
    @property
    def language(self):
        lang = self.metadata.get('kernelspec', {}).get('language', None)
        lang = lang or self.metadata.get('language_info', {}).get('name', None)
        return lang or self.metadata.get('language_info', {}).get(
            'pygments_lexer', None
        )

    def add_author(self, name, **kwargs):
        """
        Add author to metadata
        :param name: name of the author
        :param kwargs: any additional information about the author
        """
        if 'authors' not in self.metadata:
            self.metadata['authors'] = []

        # Create author info
        author_inf = {'name': name}
        author_inf.update(kwargs)

        # add author
        self.metadata['authors'].append(author_inf)

    def set_kernelspec(self, argv, display_name, language, **kwargs):
        """
        set the kernel specs.

        See https://jupyter-client.readthedocs.io/en/stable/kernels.html#kernelspecs
        for more information

        :param argv: A list of command line arguments used to start the kernel.
        :param display_name:The kernel’s name as it should be displayed in the UI.
            Unlike the kernel name used in the API, this can contain arbitrary unicode characters.
        :param language: The name of the language of the kernel.
        :param kwargs: optional keyword arguments
        """
        # Create kernelspec
        kernelspec = {'argv': argv, 'display_name': display_name, 'language': language}
        kernelspec.update(kwargs)

        # set kernelspec
        self.metadata['kernelspec'] = kernelspec


class NotebookCellMetadata(ClassicNotebook):
    def add_tag(self, tag: str):
        """
        Add tag to cell metadata.
        :param tag: tag to add
        """
        for cell in self.iter_cells():
            cell.add_tag(tag)

    def remove_tag(self, tag: str):
        """
        remove tag to cell metadata.
        :param tag: tag to remove
        """
        for cell in self.iter_cells():
            cell.remove_tag(tag)

    def set_collapsed(self, value: bool = True):
        """
        Whether the cell’s output container should be collapsed
        :param value: boolean
        """
        self.update_cell_metadata('collapsed', value)

    def set_scrolled(self, value: Union[bool, str] = False):
        """
        Whether the cell’s output is scrolled, unscrolled, or autoscrolled
        :param value: bool or ‘auto’
        """
        self.update_cell_metadata('scrolled', value)

    def set_deletable(self, value: bool = True):
        """
        If False, prevent deletion of the cell
        :param value: boolean
        """
        self.update_cell_metadata('deletable', value)

    def set_editable(self, value: bool = True):
        """
        If False, prevent editing of the cell (by definition, this also prevents deleting the cell)
        :param value: boolean
        """
        self.update_cell_metadata('editable', value)

    def set_format(self, value: str):
        """
        The mime-type of a Raw NBConvert Cell
        :param value: ‘mime/type’
        """
        self.update_cell_metadata('format', value)

    def set_name(self, value: str):
        """
        A name for the cell. Should be unique across the notebook.
        Uniqueness must be verified outside of the json schema.
        :param value: name of the cell
        """
        # TODO: check name is unique
        self.update_cell_metadata('name', value)

    def hide_source(self, value=True):
        """
        Whether the cell’s source should be shown
        :param value: boolean
        """
        self.update_cell_metadata('jupyter', {'source_hidden': value})

    def hide_output(self, value=True):
        """
        Whether the cell’s outputs should be shown
        :param value: boolean
        """
        self.update_cell_metadata('jupyter', {'outputs_hidden': value})

    def burn_attachments(self, assets_path=None, html=True):
        import re
        from functools import partial

        from nbmanips.cell.cell_utils import (
            HTML_IMG_EXPRESSION,
            HTML_IMG_REGEX,
            MD_IMG_EXPRESSION,
            MD_IMG_REGEX,
            burn_attachment,
            get_assets_path,
        )

        assets_path = get_assets_path(self, assets_path)
        compiled_md_regex = re.compile(MD_IMG_REGEX)
        compiled_html_regex = re.compile(HTML_IMG_REGEX)

        for cell in self.select('markdown_cells').iter_cells():
            # replace markdown
            rep_func = partial(
                burn_attachment,
                cell=cell,
                assets_path=assets_path,
                expr=MD_IMG_EXPRESSION,
            )
            cell.source = compiled_md_regex.sub(rep_func, cell.get_source())

            if not html:
                continue

            # replace html
            rep_func = partial(
                burn_attachment,
                cell=cell,
                assets_path=assets_path,
                expr=HTML_IMG_EXPRESSION,
            )
            cell.source = compiled_html_regex.sub(rep_func, cell.get_source())


class ContentAnalysisMixin(NotebookBase):
    @property
    def toc(self):
        markdown_cells = self.select('is_markdown')

        toc = []
        indentation_levels = []
        for cell in markdown_cells.iter_cells():
            for element in cell.soup.select('h1, h2, h3, h4, h5, h6'):
                indentation_level = int(element.name[-1]) - 1
                indentation_levels.append(indentation_level)
                toc.append((indentation_level, element.text, cell.num))

        return toc

    def ptoc(self, width=None, index=False):
        toc = self.toc

        if not toc:
            return ''

        min_indentation = min(ind_level for ind_level, _, _ in toc)

        indented_toc = [
            ('  ' * (ind - min_indentation) + title, cell_num)
            for ind, title, cell_num in toc
        ]

        if width is None:
            max_width = shutil.get_terminal_size().columns
            max_length = max(len(x) for x, _ in indented_toc) + 7
            width = min(max_width, max_length)
        width -= 7

        wrapped_toc = [(textwrap.wrap(title, width), n) for title, n in indented_toc]

        printable_toc = []
        for title, cell_num in wrapped_toc:
            if index:
                title[0] = title[0] + ' ' * (width - len(title[0])) + f'  [{cell_num}]'
            printable_toc.extend(title)

        return '\n'.join(printable_toc)

    def show_toc(self, width=None, index=True):
        print(self.ptoc(width, index))

    def add_toc(self, pos=0, bullets=False):
        from nbmanips.cell import Cell

        toc = self.toc

        if not toc:
            raise ValueError('Could not build Table of contents. No headers found.')

        min_indentation = min(ind_level for ind_level, _, _ in toc)

        numbered_toc = []
        stack = [0, 0, 0, 0, 0, 0]
        for ind, title, _ in toc:
            for i in range(ind + 1, len(stack)):
                stack[i] = 0
            stack[ind] += 1
            numbered_toc.append((stack[ind], ind, title))

        indented_toc = [
            '  ' * (ind - min_indentation)
            + ('- ' if bullets else f'{num}. ')
            + f"[{title}](#{title.replace(' ', '-')})\n"
            for num, ind, title in numbered_toc
        ]

        toc_cell = Cell(
            {'cell_type': 'markdown', 'source': '\n'.join(indented_toc), 'metadata': {}}
        )

        self.add_cell(toc_cell, pos)



import json
import re
import shutil
import urllib.parse
import warnings
from mimetypes import guess_type
from pathlib import Path
from textwrap import wrap
from typing import Union

from nbmanips.cell.color import supports_color

try:
    import pygments
    from pygments.formatters import TerminalFormatter

    FORMATTER = TerminalFormatter()
except ImportError:
    pygments = None
    TerminalFormatter = None
    FORMATTER = None

try:
    import colorama

    colorama.init()
except ImportError:
    colorama = None


# -- Constants --
# --- Attachment Constants ---
MD_IMG_REGEX = r'!\[(?P<ALT_TEXT>.*?)]\((?P<PATH>.*?)\)'
MD_IMG_EXPRESSION = r'![{ALT_TEXT}](attachment:{attachment_name})'
HTML_IMG_REGEX = (
    r'<img\s(?P<PREFIX>.*?)'
    r'src\s*=\s*\"?(?P<PATH>(?<=\")[^\"]*(?=\")|(?:[^\"\s]|(?<=\\)\s)*[^\s\\/])\"?'
    r'(?P<SUFFIX>.*?)>'
)
HTML_IMG_EXPRESSION = r'<img {PREFIX}src="attachment:{attachment_name}"{SUFFIX}>'

# -- Styles --
styles = {
    'single': '││┌─┐└─┘',
    'double': '║║╔═╗╚═╝',
    'grid': '||+-++-+',
    'separated': '||+=++=+',
    'rounded': "||.-.'-'",
    'dots': '::......',
    'simple': '  ======',
    'copy': ('', '', '#', '-', '-', '', '', ''),
}

ANSI_ESCAPE = re.compile(r'\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])')
COLOR_SUPPORTED = supports_color()
PYGMENTS_SUPPORTED = COLOR_SUPPORTED and pygments is not None


def printable_cell(text, width=None, style='single', color=None, pygments_lexer=None):
    width = width or (shutil.get_terminal_size().columns - 1)
    style_l, style_r, style_ul, style_u, style_ur, style_dl, style_d, style_dr = styles[
        style
    ]
    space_l = ' ' if style_l else ''
    space_r = ' ' if style_r else ''

    color_start, color_end = '', ''
    if color:
        color = 'WHITE' if color is None else color.upper()
        color_start, color_end = (
            vars(colorama.Fore)[color.upper()],
            colorama.Fore.RESET,
        )

    text_width = width - len(space_l) - len(space_r) - len(style_l) - len(style_r)

    code_lines = []
    for code_line in text.split('\n'):
        code_lines.extend(wrap(code_line, text_width, tabsize=4))
    diff = [text_width - len(code_line) for code_line in code_lines]

    code = '\n'.join(code_lines)
    if pygments_lexer:
        code_lines = pygments.highlight(code, pygments_lexer, FORMATTER)[:-1].split(
            '\n'
        )

    result = [
        color_start
        + style_ul
        + style_u * (width - len(style_ul) - len(style_ur))
        + style_ur
        + color_end
    ]
    result.extend(
        [
            f'{color_start}{style_l}{color_end}'
            + space_l
            + f"{line}{' ' * d}"
            + space_r
            + f'{color_start}{style_r}{color_end}'
            for line, d in zip(code_lines, diff)
        ]
    )
    result.append(
        color_start
        + style_dl
        + style_d * (width - len(style_dl) - len(style_dr))
        + style_dr
        + color_end
    )
    return '\n'.join(line.rstrip(' ') for line in result)


def monochrome(text):
    """
    Source:
    https://stackoverflow.com/questions/14693701/how-can-i-remove-the-ansi-escape-sequences-from-a-string-in-python

    :param text: any text
    """
    return ANSI_ESCAPE.sub('', text)


def get_mime_type(path):
    return guess_type(path)[0]


def _get_output_types(output_type: Union[set, dict, str]) -> set:
    if isinstance(output_type, str):
        if '/' in output_type:
            return {output_type, output_type.split('/')[0]}
        return {output_type}

    output_types = set()
    for output in output_type:
        output_types |= _get_output_types(output)
    return output_types


def _to_html(text):
    import html

    return html.escape(text).encode('ascii', 'xmlcharrefreplace').decode('ascii')


def get_assets_path(nb, assets_path=None):
    if assets_path is None:
        assets_path = getattr(nb, '_original_path', None)
        if assets_path:
            return Path(assets_path).parent
        return Path.cwd()

    return Path(assets_path)


def burn_attachment(match, cell, assets_path: Path, expr):
    path = match.group('PATH')
    if path.startswith('attachment:'):
        return match.group(0)

    path = assets_path / urllib.parse.unquote(path)
    if not path.exists():
        path = match.group('PATH')
        warnings.warn(f"Couldn't find '{path}'")
        return match.group(0)

    match_dict = match.groupdict()
    attachment_name = match_dict.pop('PATH').replace(' ', '%20')
    cell.attach(str(path), attachment_name=attachment_name)
    return expr.format(**match_dict, attachment_name=attachment_name)


def total_size(o):
    return len(json.dumps(o).encode('utf-8'))



import re

from .notebook_base import NotebookBase
from .notebook_mixins import (
    ClassicNotebook,
    ContentAnalysisMixin,
    ExportMixin,
    NotebookCellMetadata,
    NotebookMetadata,
    SlideShowMixin,
)


def _get_regex(text, case=False, regex=False):
    if not regex:
        text = re.escape(text)

    flags = 0
    flags = (flags & ~re.IGNORECASE) if case else (flags | re.IGNORECASE)

    return re.compile(text, flags)


class Notebook(
    NotebookCellMetadata,
    SlideShowMixin,
    ClassicNotebook,
    ContentAnalysisMixin,
    ExportMixin,
    NotebookMetadata,
    NotebookBase,
):
    def search(self, text, case=False, output=False, regex=False):
        """
        Return the number of the first cell containing the given text
        :param text: a string to find in cell
        :param case: True if the search is case sensitive
        :type case: default True
        :param output: True if you want the search in the output of the cell too
        :type output: default False
        :param regex: boolean whether to use regex or not
        :return:
        """

        if not regex:
            return self.select('contains', text=text, case=case, output=output).first()

        compiled_regex = _get_regex(text, case=case, regex=regex)
        return self.select('has_match', compiled_regex, output=output).first()

    def search_all(self, text, case=False, output=False, regex=False):
        """
        Return the numbers of the cells containing the given text

        :param text: a string to find in cell
        :param case: True if the search is case sensitive
        :type case: default True
        :param output: True if you want the search in the output of the cell too
        :type output: default False
        :param regex: boolean whether to use regex or not
        :return:
        """

        if not regex:
            return self.select('contains', text=text, case=case, output=output).list()

        compiled_regex = _get_regex(text, case=case, regex=regex)
        return self.select('has_match', compiled_regex, output=output).list()

    def replace(self, old, new, count=None, case=True, regex=False):
        """
        Replace matching text in the selected cells

        :param old: a string to replace in cell
        :param new: the replacement string
        :param count:
        :param case: True if the search is case sensitive
        :type case: default True
        :param regex: boolean whether to use regex or not
        """

        compiled_regex = _get_regex(old, case=case, regex=regex)
        selection = self.select('has_match', compiled_regex)
        for n_cells, cell in enumerate(selection.iter_cells(), start=1):
            cell.source = compiled_regex.sub(new, cell.get_source())
            if count is not None and n_cells >= count:
                break


class IPYNB(Notebook):
    def __new__(cls, path, name=None):
        return Notebook.read_ipynb(path, name)


class DBC(Notebook):
    def __new__(cls, path, filename=None, encoding='utf-8', name=None):
        return Notebook.read_dbc(path, filename=filename, encoding=encoding, name=name)


class ZPLN(Notebook):
    def __new__(cls, path, name=None, encoding='utf-8'):
        return Notebook.read_zpln(path, encoding=encoding, name=name)



import os

import click

from nbmanips import Notebook
from nbmanips.cli import get_selector

__all__ = ['convert']


@click.group(help='Convert a notebook to another format')
def convert():
    pass


@convert.command(help='Exports a basic HTML document.')
@click.argument('notebook_path')
@click.option('--output', '-o', help='path to export to', default=None)
@click.option(
    '--template-name', '-t', help='the name of the template to use', default=None
)
@click.option(
    '--exclude-code-cell',
    is_flag=True,
    help='exclude code cells from all templates',
    default=False,
)
@click.option(
    '--exclude-markdown',
    is_flag=True,
    help='exclude markdown cells from all templates',
    default=False,
)
@click.option(
    '--exclude-raw',
    is_flag=True,
    help='exclude unknown cells from all templates',
    default=False,
)
@click.option(
    '--exclude-unknown',
    is_flag=True,
    help='exclude unknown cells from all templates',
    default=False,
)
@click.option(
    '--exclude-input',
    is_flag=True,
    help='exclude input prompts from all templates',
    default=False,
)
@click.option(
    '--exclude-output',
    is_flag=True,
    help='exclude output prompts from all templates',
    default=False,
)
@click.option(
    '--kwarg',
    'kwargs',
    multiple=True,
    help='any additional parameters',
    type=(str, str),
)
def html(
    notebook_path,
    output,
    template_name,
    exclude_code_cell,
    exclude_markdown,
    exclude_raw,
    exclude_unknown,
    exclude_input,
    exclude_output,
    kwargs,
):
    if output is None:
        output = os.path.splitext(notebook_path)[0] + '.html'
    nb = Notebook.read(notebook_path)
    selector = get_selector()

    nb.select(selector).to_html(
        output,
        template_name=template_name,
        exclude_code_cell=exclude_code_cell,
        exclude_markdown=exclude_markdown,
        exclude_raw=exclude_raw,
        exclude_unknown=exclude_unknown,
        exclude_input=exclude_input,
        exclude_output=exclude_output,
        **dict(kwargs),
    )


@convert.command(help='Exports to a markdown document (.md)')
@click.argument('notebook_path')
@click.option('--output', '-o', help='path to export to', default=None)
@click.option(
    '--template-name', '-t', help='the name of the template to use', default=None
)
@click.option(
    '--exclude-code-cell',
    is_flag=True,
    help='exclude code cells from all templates',
    default=False,
)
@click.option(
    '--exclude-markdown',
    is_flag=True,
    help='exclude markdown cells from all templates',
    default=False,
)
@click.option(
    '--exclude-raw',
    is_flag=True,
    help='exclude unknown cells from all templates',
    default=False,
)
@click.option(
    '--exclude-unknown',
    is_flag=True,
    help='exclude unknown cells from all templates',
    default=False,
)
@click.option(
    '--exclude-input',
    is_flag=True,
    help='exclude input prompts from all templates',
    default=False,
)
@click.option(
    '--exclude-output',
    is_flag=True,
    help='exclude output prompts from all templates',
    default=False,
)
@click.option(
    '--kwarg',
    'kwargs',
    multiple=True,
    help='any additional parameters',
    type=(str, str),
)
def md(
    notebook_path,
    output,
    template_name,
    exclude_code_cell,
    exclude_markdown,
    exclude_raw,
    exclude_unknown,
    exclude_input,
    exclude_output,
    kwargs,
):
    if output is None:
        output = os.path.splitext(notebook_path)[0] + '.md'
    nb = Notebook.read(notebook_path)
    selector = get_selector()

    nb.select(selector).to_md(
        output,
        template_name=template_name,
        exclude_code_cell=exclude_code_cell,
        exclude_markdown=exclude_markdown,
        exclude_raw=exclude_raw,
        exclude_unknown=exclude_unknown,
        exclude_input=exclude_input,
        exclude_output=exclude_output,
        **dict(kwargs),
    )


@convert.command(help='Exports a Python code file.')
@click.argument('notebook_path')
@click.option('--output', '-o', help='path to export to', default=None)
@click.option(
    '--template-name', '-t', help='the name of the template to use', default=None
)
@click.option(
    '--kwarg',
    'kwargs',
    multiple=True,
    help='any additional parameters',
    type=(str, str),
)
def py(notebook_path, output, template_name, kwargs):
    if output is None:
        output = os.path.splitext(notebook_path)[0] + '.py'
    nb = Notebook.read(notebook_path)
    selector = get_selector()

    nb.select(selector).to_py(output, template_name=template_name, **dict(kwargs))


@convert.command(help='Exports HTML slides with reveal.js')
@click.argument('notebook_path')
@click.option('--output', '-o', help='path to export to', default=None)
@click.option(
    '--template-name', '-t', help='the name of the template to use', default=None
)
@click.option(
    '--exclude-code-cell',
    is_flag=True,
    help='exclude code cells from all templates',
    default=False,
)
@click.option(
    '--exclude-markdown',
    is_flag=True,
    help='exclude markdown cells from all templates',
    default=False,
)
@click.option(
    '--exclude-raw',
    is_flag=True,
    help='exclude unknown cells from all templates',
    default=False,
)
@click.option(
    '--exclude-unknown',
    is_flag=True,
    help='exclude unknown cells from all templates',
    default=False,
)
@click.option(
    '--exclude-input',
    is_flag=True,
    help='exclude input prompts from all templates',
    default=False,
)
@click.option(
    '--exclude-output',
    is_flag=True,
    help='exclude output prompts from all templates',
    default=False,
)
@click.option('--theme', help='Name of the reveal.js theme to use.', default='simple')
@click.option(
    '--transition', help='Name of the reveal.js transition to use', default='slide'
)
@click.option(
    '--scroll/--no-scroll',
    type=bool,
    help='enable scrolling within each slide',
    default=True,
)
@click.option(
    '--kwarg',
    'kwargs',
    multiple=True,
    help='any additional parameters',
    type=(str, str),
)
def slides(
    notebook_path,
    output,
    template_name,
    exclude_code_cell,
    exclude_markdown,
    exclude_raw,
    exclude_unknown,
    exclude_input,
    exclude_output,
    scroll,
    transition,
    theme,
    kwargs,
):
    if output is None:
        output = os.path.splitext(notebook_path)[0] + '.slides.html'
    nb = Notebook.read(notebook_path)
    selector = get_selector()

    nb.select(selector).to_slides(
        output,
        template_name=template_name,
        exclude_code_cell=exclude_code_cell,
        exclude_markdown=exclude_markdown,
        exclude_raw=exclude_raw,
        exclude_unknown=exclude_unknown,
        exclude_input=exclude_input,
        exclude_output=exclude_output,
        scroll=scroll,
        transition=transition,
        theme=theme,
        **dict(kwargs),
    )



from copy import deepcopy
from typing import Any, Callable, Iterator, Optional

import nbformat

from nbmanips.cell import Cell
from nbmanips.selector import Selector


class NotebookBase:
    def __init__(
        self, content: Optional[dict] = None, name=None, validate=True, copy=True
    ):
        if content is None:
            content = dict(nbformat.v4.new_notebook())

        if validate:
            self.__validate(content)

        if copy:
            self.raw_nb = deepcopy(content)
        else:
            self.raw_nb = content

        self.name = name
        self._selector = Selector(None)

    def select(self, selector: Any, *args, **kwargs):
        nb = self.reset_selection()
        nb._selector = self._selector & Selector(selector, *args, **kwargs)
        return nb

    def apply(self, func: Callable[[Cell], Optional[Cell]], neg=False):
        delete_list = []
        for cell in self.iter_cells(neg):
            num = cell.num
            new_cell = func(cell)
            if new_cell is None:
                delete_list.append(num)
            else:
                self.cells[num] = new_cell.cell

        for num in reversed(delete_list):
            del self.cells[num]

    def map(self, func: Callable[[Cell], Any], neg=False):
        return list(map(func, self.iter_cells(neg)))

    def reset_selection(self):
        notebook_selection = self.__class__(
            self.raw_nb, self.name, validate=False, copy=False
        )

        # Adding Original Notebook Path if defined
        original_path = getattr(self, '_original_path', None)
        if original_path:
            notebook_selection._original_path = original_path

        return notebook_selection

    def iter_cells(self, neg=False) -> Iterator[Cell]:
        return self._selector.iter_cells(self.raw_nb, neg=neg)

    @property
    def cells(self):
        return self.raw_nb['cells']

    @property
    def metadata(self):
        return self.raw_nb['metadata']

    @property
    def used_ids(self):
        return {cell['id'] for cell in self.cells if 'id' in cell}

    def first_cell(self):
        """
        Return the first selected cell
        :return:
        """
        for cell in self.iter_cells():
            return cell

    def last_cell(self):
        """
        Return the last selected cell
        :return:
        """
        for cell in reversed(list(self.iter_cells())):
            return cell

    def list_cells(self):
        """
        Return a list of the selected cells
        :return:
        """
        return [cell for cell in self.iter_cells()]

    def add_cell(self, cell: Cell, pos=None):
        pos = len(self) if pos is None else pos

        new_id = cell.id
        while new_id in self.used_ids:
            new_id = cell.generate_id_candidate()

        cell = cell.get_copy(new_id)
        self.cells.insert(pos, cell.cell)

    def __iter__(self) -> Iterator[Cell]:
        return self.iter_cells()

    def __add__(self, other: 'NotebookBase'):
        if not isinstance(other, NotebookBase):
            return NotImplemented

        # Copying the notebook
        raw_nb = {
            key: deepcopy(value) for key, value in self.raw_nb.items() if key != 'cells'
        }

        # Creating empty Notebook
        raw_nb['cells'] = []
        new_nb = self.__class__(raw_nb, validate=False, copy=False)

        # Concatenating the notebooks
        for cell in self.list_cells() + other.list_cells():
            new_nb.add_cell(cell)

        return new_nb

    def __mul__(self, other: int):
        if not isinstance(other, int):
            return NotImplemented

        # Copying the notebook
        raw_nb = {
            key: deepcopy(value) for key, value in self.raw_nb.items() if key != 'cells'
        }

        # Creating empty Notebook
        raw_nb['cells'] = []
        new_nb = self.__class__(raw_nb, validate=False, copy=False)

        # Concatenating the notebooks
        for _ in range(other):
            for cell in self.iter_cells():
                new_nb.add_cell(cell)
        return new_nb

    def __getitem__(self, item):
        if isinstance(item, tuple):
            return self.select(*item)
        return self.select(item)

    def __len__(self):
        if self.raw_nb is None or 'cells' not in self.raw_nb:
            return 0
        return len(self.list_cells())

    def __repr__(self):
        if self.name:
            return f'<Notebook "{self.name}">'
        else:
            return '<Notebook>'

    def __str__(self):
        return '\n'.join(str(cell) for cell in self.iter_cells())

    def __and__(self, other):
        if not isinstance(other, NotebookBase):
            return NotImplemented

        if other.raw_nb is not self.raw_nb:
            raise ValueError('and operator only works with the same Notebook.')

        nb = self.reset_selection()
        nb._selector = self._selector & other._selector
        return nb

    def __or__(self, other):
        if not isinstance(other, NotebookBase):
            return NotImplemented

        if other.raw_nb is not self.raw_nb:
            raise ValueError('and operator only works with the same Notebook.')

        nb = self.reset_selection()
        nb._selector = self._selector | other._selector
        return nb

    def __invert__(self):
        nb = self.reset_selection()
        nb._selector = ~self._selector
        return nb

    @staticmethod
    def __validate(content: dict):
        if not isinstance(content, dict):
            message = (
                f"'content' must be of type 'dict': {type(content).__name__!r} given"
            )
            if isinstance(content, str):
                message += '\nUse Notebook.read(path) to read notebook from file'
            raise ValueError(message)
        nbformat.validate(nbdict=content)



"""
Copyright (c) Django Software Foundation and individual contributors.
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer in the
       documentation and/or other materials provided with the distribution.

    3. Neither the name of Django nor the names of its contributors may be used
       to endorse or promote products derived from this software without
       specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

import os
import sys

try:
    import colorama

    colorama.init()
except (ImportError, OSError):
    HAS_COLORAMA = False
else:
    HAS_COLORAMA = True


def supports_color():
    """
    Return True if the running system's terminal supports color,
    and False otherwise.
    """

    def vt_codes_enabled_in_windows_registry():
        """
        Check the Windows Registry to see if VT code handling has been enabled
        by default, see https://superuser.com/a/1300251/447564.
        """
        try:
            # winreg is only available on Windows.
            import winreg
        except ImportError:
            return False
        else:
            reg_key = winreg.OpenKey(winreg.HKEY_CURRENT_USER, 'Console')
            try:
                reg_key_value, _ = winreg.QueryValueEx(reg_key, 'VirtualTerminalLevel')
            except FileNotFoundError:
                return False
            else:
                return reg_key_value == 1

    # isatty is not always implemented, #6223.
    is_a_tty = hasattr(sys.stdout, 'isatty') and sys.stdout.isatty()

    return is_a_tty and (
        sys.platform != 'win32'
        or HAS_COLORAMA
        or 'ANSICON' in os.environ
        or
        # Windows Terminal supports VT codes.
        'WT_SESSION' in os.environ
        or
        # Microsoft Visual Studio Code's built-in terminal supports colors.
        os.environ.get('TERM_PROGRAM') == 'vscode'
        or vt_codes_enabled_in_windows_registry()
    )



import re
from typing import Callable, ClassVar, Dict, Optional, Union

from nbmanips.cell import Cell, MarkdownCell

from .callable_selector import CallableSelector


class DefaultSelector(CallableSelector):
    default_selectors: ClassVar[Dict[str, Callable]] = {}

    def __init__(self, selector: str, *args, **kwargs):
        # TODO: use signature ?
        callable_selector = self.default_selectors[selector]
        super(DefaultSelector, self).__init__(callable_selector, *args, **kwargs)

    @classmethod
    def register_selector(cls, key: str, selector: Callable[..., bool]) -> None:
        cls.default_selectors[key] = selector


# -- Default Selectors --
def contains(
    cell: Cell, text, case=True, output=False, regex=False, flags: int = 0
) -> bool:
    """
    Selects Cells containing a certain text.

    :param cell: Cell object to select
    :param text: a string to find in cell
    :param case: True if the search is case sensitive
    :type case: default True
    :param output: True if you want the search in the output of the cell too
    :type output: default False
    :param regex: boolean whether to use regex or not
    :param flags: flags used if regex is set to True
    :type flags: int
    :return: a bool object (True if cell should be selected)
    """
    return cell.contains(text, case=case, output=output, regex=regex, flags=flags)


def has_match(cell: Cell, regex, output=False) -> bool:
    """
    Selects Cells that match a certain regex.

    :param cell: Cell object to select
    :param text: a string to find in cell
    :param case: True if the search is case sensitive
    :type case: default True
    :param output: True if you want the search in the output of the cell too
    :type output: default False
    :param regex: boolean whether to use regex or not
    :return: a bool object (True if cell should be selected)
    """

    if isinstance(regex, str):
        regex = re.compile(regex)

    return cell.has_match(regex, output=output)


def has_type(cell: Cell, type) -> bool:
    """
    Selects cells with the given type

    :param cell: Cell object to select
    :param type:
    :return: a bool object (True if cell should be selected)
    """
    return cell.type == type


def is_code(cell: Cell) -> bool:
    """
    Selects code cells

    :param cell: Cell object to select
    :return: a bool object (True if cell should be selected)
    """
    return has_type(cell, 'code')


def is_markdown(cell: Cell) -> bool:
    """
    Selects markdown cells

    :param cell: Cell object to select
    :return: a bool object (True if cell should be selected)
    """
    return has_type(cell, 'markdown')


def is_raw(cell: Cell) -> bool:
    """
    Selects raw cells

    :param cell: Cell object to select
    :return: a bool object (True if cell should be selected)
    """
    return has_type(cell, 'raw')


def has_output(cell: Cell, value: bool = True) -> bool:
    """
    Checks if the cell has any output

    :param cell: Cell object to select
    :param value: set to False if you want to select cells with no output
    :return: a bool object (True if cell should be selected)
    """
    return (cell.output != '') == value


def has_output_type(cell: Cell, output_type: Union[set, str]) -> bool:
    """
    Selects cells that have a given output_type

    :param cell: Cell object to select
    :param output_type: Output Type(MIME type) to select: text/plain, text/html, image/png, ...
    :type output_type: str
    :return: a bool object (True if cell should be selected)
    """
    if isinstance(output_type, str):
        output_types = {output_type}
    else:
        output_types = set(output_type)

    return cell.has_output_type(output_types)


def is_empty(cell: Cell) -> bool:
    """
    Selects empty cells

    :param cell: Cell object to select
    :return: a bool object (True if cell should be selected)
    """
    return cell.source == '' and has_output(cell, False)


def has_byte_size(
    cell: Cell,
    min_size=0,
    max_size: Optional[int] = None,
    output_types=None,
    ignore_source=False,
) -> bool:
    """
    Selects cells with byte size less than max_size and more than min_size.

    :param cell: Cell object to select
    :param min_size: int representing the minimum size
    :param max_size: int representing the maximum size
    :param output_types: Output Types(MIME type) to select: text/plain, text/html, image/png, ...
    :type output_types: set
    :param ignore_source: True if you want to get the size of the output only
    :return: a bool object (True if cell should be selected)
    """
    if isinstance(output_types, str):
        output_types = {output_types}

    size = cell.byte_size(output_types, ignore_source)

    return size >= min_size and (max_size is None or size < max_size)


def has_slide_type(cell: Cell, slide_type) -> bool:
    """
    Selects markdown cells that have a given slide type

    :param cell: Cell object to select
    :param slide_type: Slide Type(s): '-', 'skip', 'slide', 'subslide', 'fragment', 'notes'
    :type slide_type: str / set / list
    :return: a bool object (True if cell should be selected)
    """
    if isinstance(slide_type, str):
        slide_type = {slide_type}

    return all(
        f(cell)
        for f in [
            lambda c: 'slideshow' in c.metadata,
            lambda c: 'slide_type' in c.metadata['slideshow'],
            lambda c: c.metadata['slideshow']['slide_type'] in slide_type,
        ]
    )


def has_tag(cell: Cell, tag: str, case=False) -> bool:
    """
    Selects cells that have a certain tag

    :param cell: Cell object to select
    :param tag:
    :type tag: str
    :param case:
    :type case: bool
    :return: a bool object (True if cell should be selected)
    """
    if case:
        return tag in cell.metadata.get('tags', {})
    else:
        return tag.lower() in {
            cell_tag.lower() for cell_tag in cell.metadata.get('tags', {})
        }


def with_css_selector(cell: MarkdownCell, css_selector: str) -> bool:
    """
    Select cells that match a certain CSS Selector

    :param cell: Cell object to select
    :param css_selector: Css selector
    :type css_selector: str
    :return: a bool object (True if cell should be selected)
    """
    if not is_markdown(cell):
        return False

    return cell.soup.select_one(css_selector) is not None


def is_new_slide(cell: Cell, subslide=True) -> bool:
    """
    Selects cells where a new slide/subslide starts
    :param cell: Cell object to select
    :param subslide: False if subslides should not be selected
    :return: a bool object (True if cell should be selected)
    """
    slide_types = {'slide'}
    if subslide:
        slide_types.add('subslide')
    return has_slide_type(cell, slide_types)


# -- Default Selectors --
DefaultSelector.register_selector('contains', contains)
DefaultSelector.register_selector('has_match', has_match)
DefaultSelector.register_selector('empty', is_empty)
DefaultSelector.register_selector('is_empty', is_empty)
DefaultSelector.register_selector('has_byte_size', has_byte_size)
DefaultSelector.register_selector('has_tag', has_tag)

# -- Code Specific Selectors --
DefaultSelector.register_selector('has_output', has_output)
DefaultSelector.register_selector('has_output_type', has_output_type)

# -- Markdown Specific Selectors --
DefaultSelector.register_selector('has_html_tag', with_css_selector)
DefaultSelector.register_selector('with_css_selector', with_css_selector)

# -- Cell Types --
DefaultSelector.register_selector('has_type', has_type)
DefaultSelector.register_selector('raw_cells', is_raw)
DefaultSelector.register_selector('is_raw', is_raw)
DefaultSelector.register_selector('markdown_cells', is_markdown)
DefaultSelector.register_selector('is_markdown', is_markdown)
DefaultSelector.register_selector('code_cells', is_code)
DefaultSelector.register_selector('is_code', is_code)

# -- Slide cells --
DefaultSelector.register_selector('has_slide_type', has_slide_type)
DefaultSelector.register_selector('is_new_slide', is_new_slide)



from functools import wraps


def partial(func, *args, **keywords):
    @wraps(func)
    def new_func(*f_args, **f_keywords):
        new_keywords = {**f_keywords, **keywords}
        return func(*f_args, *args, **new_keywords)

    return new_func



from functools import partial
from typing import Callable

from ..cell import Cell
from . import SelectorBase


class IndexSelector(SelectorBase):
    def __init__(self, index: int):
        self._index = index
        super().__init__()

    def get_callable(self, nb: dict) -> Callable:
        index = self._index
        if index < 0:
            index = len(nb) + index
        return partial(self.selector, index=index)

    @classmethod
    def selector(cls, cell: Cell, index):
        return cell.num == index



import json
import os
import zipfile
from io import StringIO
from pathlib import Path
from typing import Tuple

import nbformat
from html2text import html2text

try:
    import pandas as pd
except ImportError:
    pd = None

# -- Constants --
# --- READERS Constants ---
ZPLN_PREFIXES = {
    'python': {'%python', '%pyspark', '%spark.pyspark'},
}


def get_ipynb_name(path: str) -> str:
    return os.path.splitext(os.path.basename(path))[0]


def _get_nb_from_dict(nb_dict, as_version):
    (major, minor) = nbformat.reader.get_version(nb_dict)
    nb = nbformat.versions[major].to_notebook_json(nb_dict, minor=minor)
    return nbformat.convert(nb, as_version)


def read_ipynb(notebook_path: str, version=4) -> dict:
    s = Path(notebook_path).read_text(encoding='utf-8')
    nb = nbformat.reader.reads(s)
    nb = nbformat.convert(nb, version)
    return dict(nb)


def read_zpln(notebook_path: str, version=4, encoding='utf-8'):
    with open(notebook_path, 'r', encoding=encoding) as f:
        zep_nb = json.load(f)
    name = zep_nb.get('name', os.path.splitext(os.path.basename(notebook_path))[0])
    language = zep_nb.get('defaultInterpreterGroup', 'python')
    language_prefixes = ZPLN_PREFIXES.get(language, {'%' + language})
    notebook = {
        'metadata': {'language_info': {'name': language}},
        'nbformat': 4,
        'nbformat_minor': 4,
        'cells': [],
    }

    for paragraph in zep_nb.get('paragraphs', []):
        paragraph_config = paragraph.get('config', {})
        source_hidden = paragraph_config.get('editorHide', False)
        outputs_hidden = paragraph_config.get('tableHide', False)
        cell = {
            'metadata': {
                'collapsed': source_hidden and outputs_hidden,
                'jupyter': dict(
                    source_hidden=source_hidden, outputs_hidden=outputs_hidden
                ),
            }
        }
        title = paragraph.get('title', None)
        if title:
            cell['metadata']['name'] = title

        source = paragraph.get('text', '')
        prefix = source.split()[0].strip() if source.startswith('%') else None
        if prefix:
            suffix = '\n'.join(source.split('\n')[1:])
        else:
            suffix = source

        if prefix is not None and prefix.startswith('%md'):
            cell['source'] = suffix
            cell['cell_type'] = 'markdown'
            notebook['cells'].append(cell)
            continue

        if prefix is None or prefix in language_prefixes:
            cell['source'] = suffix
        else:
            cell['source'] = source

        cell['cell_type'] = 'code'
        cell['outputs'] = []
        if (
            paragraph.get('results', None)
            and paragraph['results'].get('code', None).upper() != 'ERROR'
        ):
            for result in paragraph['results'].get('msg', []):
                result_type = result.get('type', 'TEXT').upper()
                if result_type == 'TEXT':
                    data = result.get('data', '')
                    cell['outputs'].append(
                        {'output_type': 'stream', 'text': data, 'name': 'stdout'}
                    )
                elif result_type == 'TABLE':
                    data = result.get('data', '')
                    if pd is None:
                        cell['outputs'].append(
                            {'output_type': 'stream', 'text': data, 'name': 'stdout'}
                        )
                    else:
                        data = pd.read_csv(StringIO(data), sep='\t').to_html()
                        cell['outputs'].append(
                            {
                                'output_type': 'display_data',
                                'data': {'text/html': data},
                                'metadata': {},
                            }
                        )
                elif result_type in 'HTML':
                    data = result.get('data', '')
                    cell['outputs'].append(
                        {
                            'output_type': 'display_data',
                            'data': {'text/html': data},
                            'metadata': {},
                        }
                    )

        cell['execution_count'] = None

        notebook['cells'].append(cell)

    nb_node = _get_nb_from_dict(notebook, as_version=version)
    return name, dict(nb_node)


def read_dbc(
    notebook_path: str, version=4, filename=None, encoding='utf-8'
) -> Tuple[str, dict]:

    if zipfile.is_zipfile(notebook_path):
        with zipfile.ZipFile(notebook_path, 'r') as zf:
            if filename is None:
                names = zf.namelist()
                files = [name for name in names if not name.endswith('/')]
                if len(files) > 1:
                    raise ValueError(
                        f'Multiple Notebooks in archive: {files}\nSpecify the notebook filename.'
                    )
                filename = files[0]

            dbc_nb = json.loads(zf.read(filename).decode(encoding))
    else:
        if filename is not None or filename != os.path.basename(notebook_path):
            raise ValueError(f'Invalid filename: {filename}')
        with open(notebook_path, 'r', encoding=encoding) as f:
            dbc_nb = json.load(f)

    name = dbc_nb.get('name', os.path.splitext(os.path.basename(notebook_path))[0])
    language = dbc_nb.get(
        'language', os.path.splitext(os.path.basename(notebook_path))[1]
    )
    language_prefix = f'%{language}' if language != 'python' else '%py'
    notebook = {
        'metadata': {'language_info': {'name': language}},
        'nbformat': 4,
        'nbformat_minor': 4,
        'cells': [],
    }

    for command in dbc_nb.get('commands', []):
        cell = {'metadata': {'collapsed': command.get('collapsed', False)}}
        source = command.get('command', '')
        prefix = source.split('\n')[0].strip() if source.startswith('%') else None
        if prefix:
            suffix = '\n'.join(source.split('\n')[1:])
        else:
            suffix = source

        if prefix is not None and prefix.startswith('%md'):
            cell['source'] = suffix
            cell['cell_type'] = 'markdown'
            notebook['cells'].append(cell)
            continue

        if prefix is None or prefix == language_prefix:
            cell['source'] = suffix
        else:
            cell['source'] = source

        cell['cell_type'] = 'code'
        cell['outputs'] = []
        if (
            command.get('results', None)
            and command['results'].get('type', None) == 'html'
        ):
            html = command['results'].get('data', '')
            cell['outputs'].append(
                {
                    'output_type': 'display_data',
                    'data': {'text/html': html},
                    'metadata': {},
                }
            )

        error_summary = html2text(command.get('errorSummary', None) or '')
        error = html2text(command.get('error', None) or '')
        if error_summary or error:
            if ':' in error_summary:
                ename, evalue = error_summary.split(':', 1)
            else:
                ename, evalue = '', error_summary
            cell['outputs'].append(
                {
                    'output_type': 'error',
                    'ename': ename,
                    'evalue': evalue,
                    'traceback': error.split('\n'),
                }
            )

        cell['execution_count'] = None

        notebook['cells'].append(cell)

    nb_node = _get_nb_from_dict(notebook, as_version=version)
    return name, dict(nb_node)


def write_ipynb(nb_dict: dict, notebook_path: str, version=nbformat.NO_CONVERT) -> None:
    nb_node = dict_to_ipynb(nb_dict)
    nbformat.write(nb_node, notebook_path, version)


def dict_to_ipynb(nb_dict: dict, default_version=4) -> nbformat.NotebookNode:
    version = nb_dict.get('nbformat', default_version)
    return _get_nb_from_dict(nb_dict, as_version=version)



import click
import colorama

from nbmanips import Notebook
from nbmanips.cell.cell_utils import styles
from nbmanips.cli import get_selector

_COLORS = list(set(vars(colorama.Fore)) - {'RESET'})

__all__ = ['show', 'count', 'first', 'last', 'list_', 'search', 'toc']


@click.command(help='show notebook in human readable format')
@click.argument('notebook_path')
@click.option('--width', '-w', type=int, default=None)
@click.option('--output/--no-output', '-o/-no', type=bool, default=True)
@click.option('--exclude-output-type', '-e', 'excluded_data_types', multiple=True)
@click.option('--pygments/--no-pygments', '-p/-np', type=bool, default=None)
@click.option(
    '--style',
    '-s',
    type=click.Choice(styles.keys(), case_sensitive=False),
    default='single',
)
@click.option(
    '--border-color',
    '-bc',
    type=click.Choice(_COLORS, case_sensitive=False),
    default=None,
)
@click.option('--image-width', '-iw', type=int, default=None)
@click.option('--image-color/--no-image-color', '-ic/-nic', type=bool, default=None)
@click.option('--parser', '-p', 'parsers', type=str, multiple=True)
@click.option('--truncate', '-t', type=int, default=None)
def show(
    notebook_path,
    width,
    pygments,
    output,
    style,
    border_color,
    parsers,
    image_width,
    image_color,
    excluded_data_types,
    truncate,
):
    nb = Notebook.read(notebook_path)
    selector = get_selector()

    parsers_config = None
    if image_width or image_color:
        parsers_config = {'image': {'width': image_width, 'colorful': image_color}}

    # image_color, image_width
    nb.select(selector).show(
        width,
        exclude_output=not output,
        use_pygments=pygments,
        style=style,
        border_color=border_color,
        parsers=parsers or None,
        parsers_config=parsers_config,
        excluded_data_types=excluded_data_types or None,
        truncate=truncate,
    )


@click.command(help='count selected cells')
@click.argument('notebook_path')
def count(notebook_path):
    nb = Notebook.read(notebook_path)
    selector = get_selector()

    result = nb.select(selector).count()
    click.echo(result)


@click.command(help='Return the number of the first selected cell')
@click.argument('notebook_path')
def first(notebook_path):
    nb = Notebook.read(notebook_path)
    selector = get_selector()

    result = nb.select(selector).first()
    click.echo(result)


@click.command(help='Return the number of the last selected cell')
@click.argument('notebook_path')
def last(notebook_path):
    nb = Notebook.read(notebook_path)
    selector = get_selector()

    result = nb.select(selector).last()
    click.echo(result)


@click.command(help='Return the numbers of the selected cells')
@click.argument('notebook_path')
def list_(notebook_path):
    nb = Notebook.read(notebook_path)
    selector = get_selector()

    result = nb.select(selector).list()
    click.echo(result)


@click.command(help='Search string in all selected cells')
@click.argument('notebook_path')
@click.option('--text', '-t', required=True)
@click.option('--case/--no-case', default=False)
@click.option('--regex', '-r', is_flag=True, default=False)
@click.option('--output', '-o', is_flag=True, default=False)
def search(notebook_path, text, case, output, regex):
    nb = Notebook.read(notebook_path)
    selector = get_selector()

    result = nb.select(selector).search_all(text, case, output, regex)
    click.echo(result)


@click.command(help='Return the numbers of the selected cells')
@click.argument('notebook_path')
@click.option('--width', '-w', type=int, required=False, default=None)
@click.option('--index/--no-index', '-i/-ni', is_flag=True, default=True)
def toc(notebook_path, width, index):
    nb = Notebook.read(notebook_path)
    selector = get_selector()

    result = nb.select(selector).ptoc(width, index=index)
    click.echo(result)



import warnings
from pathlib import Path

import click
import cloudpickle


def export(nb, input_path, output_path, force=False):
    default_output = output_path is None
    output_path = input_path if output_path is None else output_path
    if not force and Path(output_path).exists():
        click.echo(
            f'Notebook "{output_path}" already exists.' ' Use --force to overwrite'
        )
        raise click.Abort()

    if output_path.lower().endswith('.dbc'):
        return nb.to_dbc(output_path)

    if output_path.lower().endswith('.zpln'):
        if default_output:
            output_path = Path(output_path).resolve()
            root_path = output_path.parent
            output_path = str(root_path / (root_path.stem + '.ipynb'))
        else:
            warnings.warn('Zeppelin Notebooks exports are not supported.')

    nb.to_ipynb(output_path)


def get_selector():
    binary_stream = click.get_binary_stream('stdin')
    if binary_stream.isatty():
        return None

    stream = binary_stream.read()
    if not stream:
        return None

    return cloudpickle.loads(stream)



from abc import ABCMeta, abstractmethod

from nbmanips.cell.cell_utils import COLOR_SUPPORTED

try:
    from html2text import html2text
except ImportError:
    html2text = None

try:
    from img2text import img_to_ascii
except ImportError:
    img_to_ascii = None


class ParserBase(metaclass=ABCMeta):
    @abstractmethod
    def parse(self, content, **kwargs):
        return content

    @property
    def default_state(self):
        return True


class TextParser(ParserBase):
    def parse(self, content, **kwargs):
        return content


class ImageParser(ParserBase):
    def parse(
        self,
        content,
        width=80,
        colorful=COLOR_SUPPORTED,
        bright=COLOR_SUPPORTED,
        reverse=True,
        **kwargs,
    ):
        if callable(img_to_ascii):
            return img_to_ascii(
                content,
                base64=True,
                colorful=colorful,
                reverse=reverse,
                width=width,
                bright=bright,
                **kwargs,
            )
        else:
            raise ModuleNotFoundError(
                'You need to pip install img2text for readable option'
            )

    @property
    def default_state(self):
        return img_to_ascii is not None


class HtmlParser(ParserBase):
    def parse(self, content, width=78, **kwargs):
        if callable(html2text):
            return html2text(content, bodywidth=width, **kwargs)
        else:
            raise ModuleNotFoundError(
                'You need to pip install html2txt for readable option'
            )

    @property
    def default_state(self):
        return html2text is not None



import os.path
from functools import reduce
from operator import add

import click

from nbmanips import Notebook
from nbmanips.cli import export, get_selector

__all__ = [
    'erase',
    'delete',
    'keep',
    'replace',
    'auto_slide',
    'erase_output',
    'split',
    'burn',
]


@click.command(help='Erase the content of the selected cells')
@click.argument('notebook_path')
@click.option('--output', '-o', default=None, type=str)
@click.option(
    '--force',
    '-f',
    is_flag=True,
    default=False,
    help='Do not prompt for confirmation if file already exists',
)
def erase(notebook_path, output, force):
    nb = Notebook.read(notebook_path)
    selector = get_selector()

    nb.select(selector).erase()
    export(nb, notebook_path, output, force=force)


@click.command(help='Delete the selected cells')
@click.argument('notebook_path')
@click.option('--output', '-o', default=None)
@click.option(
    '--force',
    '-f',
    is_flag=True,
    default=False,
    help='Do not prompt for confirmation if file already exists',
)
def delete(notebook_path, output, force):
    nb = Notebook.read(notebook_path)
    selector = get_selector()

    nb.select(selector).delete()
    export(nb, notebook_path, output, force=force)


@click.command(help='Delete all the non-selected cells')
@click.argument('notebook_path')
@click.option('--output', '-o', default=None)
@click.option(
    '--force',
    '-f',
    is_flag=True,
    default=False,
    help='Do not prompt for confirmation if file already exists',
)
def keep(notebook_path, output, force):
    nb = Notebook.read(notebook_path)
    selector = get_selector()

    nb.select(selector).keep()
    export(nb, notebook_path, output, force=force)


@click.command(help='replace string in all selected cells')
@click.argument('notebook_path')
@click.option('--output', '-o', default=None)
@click.option('--old', '-t', required=True)
@click.option('--new', '-n', required=True)
@click.option('--count', '--max', '-m', 'count_', type=int, default=None)
@click.option('--regex', '-r', is_flag=True, default=False)
@click.option('--case/--no-case', '-c/-nc', default=True)
@click.option(
    '--force',
    '-f',
    is_flag=True,
    default=False,
    help='Do not prompt for confirmation if file already exists',
)
def replace(notebook_path, output, old, new, case, count_, regex, force):
    nb = Notebook.read(notebook_path)
    selector = get_selector()

    nb.select(selector).replace(old, new, count_, case, regex)
    export(nb, notebook_path, output, force=force)


@click.command(help='replace string in all selected cells')
@click.argument('notebook_path')
@click.option('--output', '-o', default=None)
@click.option('--max-cells', type=int, default=3)
@click.option('--max-images', type=int, default=1)
@click.option('--delete-empty/--keep-empty', 'delete_empty', default=True)
@click.option(
    '--force',
    '-f',
    is_flag=True,
    default=False,
    help='Do not prompt for confirmation if file already exists',
)
def auto_slide(notebook_path, output, max_cells, max_images, delete_empty, force):
    nb = Notebook.read(notebook_path)
    selector = get_selector()

    nb.select(selector).auto_slide(max_cells, max_images, delete_empty=delete_empty)
    export(nb, notebook_path, output, force=force)


@click.command(help='Erase the output content of the selected cells')
@click.argument('notebook_path')
@click.option('--output', '-o', default=None)
@click.option('--output-type', 'output_types', multiple=True)
@click.option(
    '--force',
    '-f',
    is_flag=True,
    default=False,
    help='Do not prompt for confirmation if file already exists',
)
def erase_output(notebook_path, output, output_types, force):
    nb = Notebook.read(notebook_path)
    selector = get_selector()

    if output_types:
        output_types = set(output_types)
    else:
        output_types = None

    nb.select(selector).erase_output(output_types)
    export(nb, notebook_path, output, force=force)


@click.command(help='Split the notebook based the cell indexes')
@click.argument('notebook_path')
@click.argument('indexes', nargs=-1, required=False)
@click.option('--output', '-o', default=None)
@click.option('--index', '-i', multiple=True)
@click.option('--use-selection', '-s', is_flag=True, default=False)
@click.option(
    '--force',
    '-f',
    is_flag=True,
    default=False,
    help='Do not prompt for confirmation if file already exists',
)
def split(notebook_path, output, indexes, index, force, use_selection):
    if index or indexes:
        indexes = reduce(
            add, [index.split(',') for index in list(indexes) + list(index)]
        )
        indexes = [int(index) for index in indexes]
    elif not use_selection:
        raise ValueError('You need to specify the cells to split on')

    if indexes and use_selection:
        raise ValueError('Cannot use selection and indexes at the same time')

    nb = Notebook.read(notebook_path)
    selector = get_selector()

    if use_selection:
        nbs = nb.select(selector).split_on_selection()
    else:
        nbs = nb.select(selector).split(*indexes)

    # Exporting
    base, ext = os.path.splitext(notebook_path)
    input_path = base + '-%d' + ext
    for i, nb in enumerate(nbs):
        if output:
            output_path = output % i
        else:
            output_path = None
        export(nb, input_path % i, output_path, force=force)


@click.command(help='Burn the images in markdown cells as attachments')
@click.argument('notebook_path')
@click.option(
    '--assets-path',
    '-a',
    type=str,
    default=None,
    help='Path of the assets used in the notebook',
)
@click.option(
    '--html/--no-html',
    is_flag=True,
    default=True,
    help='burn images in img html tags as attachments',
)
@click.option('--output', '-o', default=None)
@click.option(
    '--force',
    '-f',
    is_flag=True,
    default=False,
    help='Do not prompt for confirmation if file already exists',
)
def burn(notebook_path, assets_path, output, force, html):
    nb: Notebook = Notebook.read(notebook_path)
    selector = get_selector()

    nb.select(selector).burn_attachments(assets_path=assets_path, html=html)
    export(nb, notebook_path, output, force=force)



from pathlib import Path

import setuptools

long_description = Path('README.md').read_text(encoding='utf-8')
version = Path('nbmanips/VERSION').read_text(encoding='utf-8').strip()


setuptools.setup(
    name='nbmanips',
    version=version,
    author='Dhia HMILA',
    author_email='dhiahmila@gmail.com',
    description='nbmanips allows you easily manipulate notebook files',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='MIT',
    url='https://github.com/hmiladhia/nbmanips',
    packages=setuptools.find_packages(include=[('nbmanips*')]),
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'nbmanips=nbmanips.__main__:nbmanips',
            'nb=nbmanips.__main__:nbmanips',
        ],
    },
    install_requires=[
        'nbconvert>=6.0.0',
        'nbformat>=5.1.3',
        'html2text==2020.1.16',
        'cloudpickle>=1.6.*',
        'click>=7.1.*',
        'Pygments>=2.10.*',
        'colorama>=0.4.*',
        'beautifulsoup4>=4.10.*',
        'lxml',
    ],
    extras_require={
        'images': ['img2text>=0.0.2'],
    },
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
    keywords=[
        'jupyter',
        'notebook',
        'ipynb',
        'dbc',
        'zpln',
        'zeppelin',
        'slides',
        'notebooks',
    ],
)



from abc import ABC, abstractmethod
from copy import copy
from itertools import filterfalse
from typing import Callable, Iterator, List, Tuple

from nbmanips._utils import partial
from nbmanips.cell import Cell


class SelectorBase(ABC):
    def __init__(self):
        self._neg = False

    @abstractmethod
    def get_callable(self, nb: dict) -> Callable[..., bool]:
        pass

    def iter_cells(self, nb, neg=False) -> Iterator[Cell]:
        selector = self.get_callable(nb)
        filter_method = filterfalse if (self._neg ^ neg) else filter
        return filter_method(
            selector, (Cell(cell, i) for i, cell in enumerate(nb['cells']))
        )

    def __invert__(self):
        selector = copy(self)
        selector._neg = not selector._neg
        return selector

    def __and__(self, other: 'SelectorBase'):
        if not isinstance(other, SelectorBase):
            return NotImplemented

        if isinstance(other, TrueSelector):
            return other & self

        return ListSelector([self]) & other

    def __or__(self, other: 'SelectorBase'):
        if not isinstance(other, SelectorBase):
            return NotImplemented

        if isinstance(other, TrueSelector):
            return other & self

        return ListSelector([self], type='or') | other


class TrueSelector(SelectorBase):
    def iter_cells(self, nb, neg=False) -> Iterator[Cell]:
        if self._neg ^ neg:
            return (_ for _ in range(0))
        return (Cell(cell, i) for i, cell in enumerate(nb['cells']))

    def get_callable(self, nb):
        return lambda cell: True

    def __and__(self, other: 'SelectorBase'):
        if not isinstance(other, SelectorBase):
            return NotImplemented

        if self._neg:
            return copy(self)
        return copy(other)

    def __or__(self, other: 'SelectorBase'):
        if not isinstance(other, SelectorBase):
            return NotImplemented

        if self._neg:
            return copy(other)
        return copy(self)


class ListSelector(SelectorBase):
    def __init__(self, selector, *args, **kwargs):
        from . import Selector

        selector = list(selector)
        if len(args) == len(selector):
            args_kwargs_list = args
        else:
            args_kwargs_list = tuple(args[0]) if args else tuple({} for _ in selector)

        # Parsing args
        args_list, kwargs_list = self.__parse_list_args(args_kwargs_list)

        # Creating list of selectors
        self._and = self._check_sanity(kwargs)
        self._list: List[SelectorBase] = [
            Selector(sel, *args, **kwargs)
            for sel, args, kwargs in zip(selector, args_list, kwargs_list)
        ]
        super().__init__()

    def __and__(self, other: SelectorBase):
        if not isinstance(other, SelectorBase):
            return NotImplemented

        if not self._and:
            return super().__and__(other)

        selector = copy(self)
        selector._list = copy(selector._list)
        if isinstance(other, ListSelector) and (other._and == self._and):
            selector._list.extend(other._list)
        else:
            selector._list.append(other)
        return selector

    def __or__(self, other: SelectorBase):
        if not isinstance(other, SelectorBase):
            return NotImplemented

        if self._and:
            return super().__or__(other)

        selector = copy(self)
        selector._list = copy(selector._list)
        if isinstance(other, ListSelector) and (other._and == self._and):
            selector._list.extend(other._list)
        else:
            selector._list.append(other)
        return selector

    def get_callable(self, nb: dict) -> Callable:
        op = all if self._and else any
        return partial(self.__combine, op=op, nb=nb)

    def __combine(self, cell: Cell, op, nb):
        return op((sel._neg ^ sel.get_callable(nb)(cell)) for sel in self._list)

    @staticmethod
    def _check_sanity(kwargs):
        _type = kwargs.pop('type', 'and')
        if _type not in {'and', 'or'}:
            raise ValueError('type can only have the following values: {"and", "or"}')

        if kwargs:
            raise ValueError("only 'type' keyword is allowed as keyword argument")

        return _type == 'and'

    @staticmethod
    def __parse_list_args(list_args: tuple) -> Tuple[list, list]:
        args_list: list = []
        kwargs_list: list = []
        for arg in list_args:
            if isinstance(arg, dict):
                args_list.append([])
                kwargs_list.append(arg)
            elif (
                isinstance(arg, tuple)
                and len(arg) == 2
                and hasattr(arg[0], '__iter__')
                and isinstance(arg[1], dict)
            ):
                args_list.append(arg[0])
                kwargs_list.append(arg[1])
            elif hasattr(arg, '__iter__'):
                args_list.append(arg)
                kwargs_list.append({})
            else:
                raise ValueError('Cannot parse arguments:', str(arg))
        return args_list, kwargs_list



import re
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

try:
    import pygments
    from pygments.lexers import get_lexer_by_name

    _MARKDOWN_LEXER = get_lexer_by_name('md')
except ImportError:
    pygments = None
    get_lexer_by_name = None

from nbmanips.cell.cell_utils import (
    FORMATTER,
    get_mime_type,
    monochrome,
    printable_cell,
    total_size,
)

from .cell_output import CellOutput


class Cell:
    _cell_types: Dict[str, type] = {}

    def __init__(self, content, num=None):
        self.cell = content
        self._num = num

    def __getitem__(self, key):
        return self.cell[key]

    def __setitem__(self, key, value):
        self.cell[key] = value

    @property
    def type(self):
        return self.cell['cell_type']

    @property
    def id(self):
        return self.cell.get('id', None)

    @id.setter
    def id(self, new_id):
        self.cell['id'] = new_id

    @property
    def num(self):
        return self._num

    @property
    def metadata(self):
        return self.cell['metadata']

    @property
    def source(self):
        return self.get_source().strip()

    @source.setter
    def source(self, source: str):
        self.set_source(source)

    @property
    def output(self):
        return self.get_output(text=True).strip()

    @property
    def outputs(self):
        return map(CellOutput, self.cell.get('outputs', []))

    def get_copy(self, new_id=None):
        from copy import deepcopy

        cell = self.__class__(deepcopy(self.cell), None)
        if new_id is not None:
            cell.id = new_id
        return cell

    def get_output(
        self, text=True, parsers=None, parsers_config=None, excluded_data_types=None
    ):
        """
        Tries its best to return a readable output from cell

        :param text:
        :param parsers: parsers to use
        :param parsers_config: custom config for parsers
        :param excluded_data_types: data types to exclude
        :return:
        """

        outputs = []
        for cell_output in self.outputs:
            text_output = cell_output.to_str(
                parsers, parsers_config, excluded_data_types
            )
            outputs.append(text_output)

        if text:
            return '\n'.join(outputs)
        return outputs

    def get_source(self, text=True):
        source = self.cell['source']

        if text and not isinstance(source, str):
            return ''.join(source)
        return source

    def set_source(self, content: Union[str, List[str]], text=False):
        if text:
            assert isinstance(content, str), 'content is not of type str if text=True'
            lines = content.split('\n')
            content = [
                f'{line}\n' if i != len(lines) else line for i, line in enumerate(lines)
            ]
        self.cell['source'] = content

    def contains(self, text, case=True, output=False, regex=False, flags=0):
        search_target = self.source
        if output:
            search_target += '\n' + self.output

        if not regex:
            if not case:
                text = text.lower()
                search_target = search_target.lower()

            return text in search_target

        if case is False:
            flags = flags | re.IGNORECASE
        else:
            flags = flags & ~re.IGNORECASE
        return bool(re.search(text, search_target, flags=flags))

    def has_match(self, regex, output=False):
        search_target = self.source
        if output:
            search_target += '\n' + self.output

        return bool(regex.search(search_target))

    def erase_output(self, output_types: Optional[Union[str, set]] = None):
        """
        erase output of cells that have a given output_type

        :param output_types: Output Type(MIME type) to delete: text/plain, text/html, image/png, ...
        :type output_types: set or str or None to delete all output
        """
        outputs = list(self.outputs)
        if len(outputs) == 0:
            return

        if output_types is None:
            self['outputs'] = []
            return
        elif isinstance(output_types, str):
            output_types = {output_types}
        else:
            output_types = set(output_types)

        new_outputs = []
        for cell_output in outputs:
            new_output = cell_output.erase_output(output_types)
            if new_output is not None:
                new_outputs.append(new_output)

        self['outputs'] = new_outputs

    def has_output_type(self, output_types: set):
        """
        Select cells that have a given output_type

        :param output_types: Output Types(MIME type) to select: text/plain, image/png, ...
        :type output_types: set
        :return: a bool object (True if cell should be selected)
        """
        return any(
            cell_output.has_output_type(output_types) for cell_output in self.outputs
        )

    def byte_size(self, output_types: Optional[set] = None, ignore_source=False):
        """
        returns the byte size of the cell.

        :param output_types: Output Types(MIME type) to select: text/plain, image/png, ...
        :type output_types: set
        :param ignore_source: True if you want to get the size of the output only
        :return: a bool object (True if cell should be selected)
        """
        size = 0 if ignore_source else total_size(self['source'])
        size += sum(cell_output.byte_size(output_types) for cell_output in self.outputs)
        return size

    def to_str(self, width=None, *args, **kwargs):
        return self.source

    def show(self):
        print(self)

    def __repr__(self):
        return f'<Cell {self.num}>' if self.num else '<Cell>'

    def __str__(self):
        return self.to_str(width=None, style='single', color=None)

    # metadata
    def update_metadata(self, key: str, value: Any):
        """
        Add metadata to the selected cells
        :param key: metadata key
        :param value: metadata value
        """
        if 'metadata' not in self.cell:
            self.cell['metadata'] = {}

        if key in self.cell['metadata'] and isinstance(
            self.cell['metadata'][key], dict
        ):
            self.metadata[key].update(value)
        else:
            self.metadata[key] = value

    def add_tag(self, tag: str):
        """
        Add tag to cell metadata.
        :param tag: tag to add
        """
        if 'metadata' not in self.cell:
            self.cell['metadata'] = {}

        if 'tags' not in self.metadata:
            self.metadata['tags'] = []

        if tag in self.metadata['tags']:
            return

        self.metadata['tags'].append(tag)

    def remove_tag(self, tag: str):
        """
        remove tag to cell metadata.
        :param tag: tag to remove
        """

        if 'metadata' not in self.cell or 'tags' not in self.metadata:
            return

        while tag in self.metadata['tags']:
            self.metadata['tags'].remove(tag)

    @staticmethod
    def generate_id_candidate():
        import uuid

        return uuid.uuid4().hex[:8]

    def __new__(cls, content, *args, **kwargs):
        cell_type = content['cell_type']
        cell_class = cls._cell_types[cell_type]
        obj = super().__new__(cell_class)
        cell_class.__init__(obj, content, *args, **kwargs)
        return obj

    def __init_subclass__(cls, cell_type=None, **kwargs):
        super().__init_subclass__(**kwargs)
        if cell_type:
            cls._cell_types[cell_type] = cls


class CodeCell(Cell, cell_type='code'):
    def to_str(
        self,
        width=None,
        style='single',
        use_pygments=None,
        pygments_lexer=None,
        color=None,
        exclude_output=False,
        parsers=None,
        parsers_config=None,
        excluded_data_types=None,
        truncate=None,
    ):
        sources = [
            printable_cell(
                self.source,
                width=width,
                style=style,
                color=color,
                pygments_lexer=pygments_lexer,
            )
        ]

        if not exclude_output:
            output = self.get_output(
                text=True,
                parsers=parsers,
                parsers_config=parsers_config,
                excluded_data_types=excluded_data_types,
            ).strip()
            monochrome_output = monochrome(output)
            if truncate is not None and len(monochrome_output) > truncate >= 0:
                output = monochrome_output[:truncate] + ' [...]'
            if output:
                sources.append(output)

        return '\n'.join(sources)


class MarkdownCell(Cell, cell_type='markdown'):
    _bs4_parser = 'lxml'

    def to_str(
        self,
        width=None,
        style='single',
        use_pygments=None,
        pygments_lexer=None,
        color=None,
        **kwargs,
    ):
        use_pygments = pygments is not None if use_pygments is None else use_pygments

        if use_pygments:
            if pygments is None:
                raise ModuleNotFoundError(
                    'You need to install pygments first.\n pip install pygments'
                )
            return pygments.highlight(self.source, _MARKDOWN_LEXER, FORMATTER)[:-1]
        else:
            return self.source

    @property
    def attachments(self):
        return self.cell.setdefault('attachments', {})

    def attach(self, path: Union[str, Path], attachment_name: Optional[str] = None):
        import base64

        mime_type = get_mime_type(str(path))
        path = Path(path)
        attachment_name = attachment_name or path.name
        self.attachments[attachment_name] = {
            mime_type: base64.encodebytes(path.read_bytes()).decode('utf-8')
        }

    @property
    def html(self):
        from nbconvert.filters.markdown_mistune import IPythonRenderer, MarkdownWithMath

        renderer = IPythonRenderer(
            escape=False, attachments=self.attachments, exclude_anchor_links=True
        )
        return MarkdownWithMath(renderer=renderer).render(self.source)

    @property
    def soup(self):
        from bs4 import BeautifulSoup

        return BeautifulSoup(self.html, self._bs4_parser)


class RawCell(Cell, cell_type='raw'):
    pass
